#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from pathlib import Path
from dataclasses import dataclass, asdict, fields, replace
from typing import Union, Optional, Dict, List, Any

import yaml
import dacite
from dacite import from_dict, Config

from isaac_toolkit.logging import get_logger

logger = get_logger()

DEFAULT_CONFIG = {
    "logging": {
        "console": {
            "level": "INFO",
        },
        "file": {
            "level": "DEBUG",
            "rotate": False,
            "limit": 1000,
        },
    },
    "memgraph": {
        "hostname": "localhost",
        "port": 7687,
        "user": "",
        "password": "",
        "database": "memgraph",
    },
    "artifacts": {
        "instr_trace": {
            "fmt": "pickle",
            "engine": None,
            "compression_method": "zstd",
            "compression_level": None,
        },
        "trace": {
            "fmt": "pickle",
            "engine": None,
            "compression_method": "zstd",
            "compression_level": None,
        },
        "table": {
            "fmt": "pickle",
            "engine": None,
            "compression_method": None,
            "compression_level": None,
        },
    },
    "flow": {
        "demo": {
            "stages": None,
            "riscv": None,
            "mlonmcu": None,
            # "llvm": None,
            "cdfg": None,
            "sim": None,
            "choose": None,
            "query": None,
            "coredsl": None,
            "hls": None,
            "asip_syn": None,
            "fpga_syn": None,
            "experiment": None,
            "docker": None,
            "service": None,
        },
    },
}


def check_supported_types(data):
    ALLOWED_TYPES = (int, float, str, bool)
    if isinstance(data, dict):
        for value in data.values():
            check_supported_types(value)
    elif isinstance(data, list):
        for x in data:
            check_supported_types(x)
    else:
        if data is not None:
            assert isinstance(data, ALLOWED_TYPES), f"Unsupported type: {type(data)}"


class YAMLSettings:
    @classmethod
    def from_dict(cls, data: dict):
        """Convert dict into instance of YAMLSettings."""
        try:
            return from_dict(data_class=cls, data=data, config=Config(strict=True))
        except dacite.exceptions.UnexpectedDataError as err:
            logger.error("Unexpected key in ISAACConfig. Check for missmatch between ISAAC Toolkit versions!")
            raise err

    @classmethod
    def from_yaml(cls, text: str):
        data = yaml.safe_load(text)
        return cls.from_dict(data)

    @classmethod
    def from_yaml_file(cls, path: Path):
        with open(path, "r") as file:
            data = yaml.safe_load(file)
        return cls.from_dict(data=data)

    def to_yaml(self):
        data = asdict(self)
        check_supported_types(data)
        text = yaml.dump(data)
        return text

    def to_yaml_file(self, path: Path):
        text = self.to_yaml()
        with open(path, "w") as file:
            file.write(text)

    def merge(self, other: "YAMLSettings", overwrite: bool = False, inplace: bool = False):
        """Merge two instances of YAMLSettings."""
        if not inplace:
            ret = replace(self)  # Make a copy of self
        for f1 in fields(other):
            k1 = f1.name
            v1 = getattr(other, k1)
            if v1 is None:
                continue
            t1 = type(v1)
            found = False
            for f2 in fields(self):
                k2 = f2.name
                v2 = getattr(self, k2)
                if k2 == k1:
                    found = True
                    if v2 is None:
                        if inplace:
                            setattr(self, k2, v1)
                        else:
                            setattr(ret, k2, v1)
                    else:
                        t2 = type(v2)
                        assert t1 is t2, "Type conflict"
                        if isinstance(v1, YAMLSettings):
                            v2.merge(v1, overwrite=overwrite, inplace=True)
                        elif isinstance(v1, dict):
                            if overwrite:
                                v2.clear()
                                v2.update(v1)
                            else:
                                for dict_key, dict_val in v1.items():
                                    if dict_key in v2:
                                        if isinstance(dict_val, YAMLSettings):
                                            assert isinstance(v2[dict_key], YAMLSettings)
                                            v2[dict_key].merge(dict_val, overwrite=overwrite, inplace=True)
                                        elif isinstance(dict_val, dict):
                                            v2[dict_key].update(dict_val)
                                        else:
                                            v2[dict_key] = dict_val
                                    else:
                                        v2[dict_key] = dict_val
                        elif isinstance(v1, list):
                            if overwrite:
                                v2.clear()
                            new = [x for x in v1 if x not in v2]
                            v2.extend(new)
                        else:
                            assert isinstance(
                                v2, (int, float, str, bool, Path)
                            ), f"Unsupported field type for merge {t1}"
                            if inplace:
                                setattr(self, k1, v1)
                            else:
                                setattr(ret, k1, v1)
                    break
            assert found
        if not inplace:
            return ret


@dataclass
class ConsoleLoggingSettings(YAMLSettings):
    level: Union[int, str] = logging.INFO


@dataclass
class FileLoggingSettings(YAMLSettings):
    level: Union[int, str] = logging.INFO
    limit: Optional[int] = None  # TODO: implement
    rotate: bool = False  # TODO: implement


@dataclass
class LoggingSettings(YAMLSettings):
    console: ConsoleLoggingSettings
    file: FileLoggingSettings


@dataclass
class MemgraphSettings(YAMLSettings):
    hostname: Optional[str] = None
    port: Optional[int] = None
    user: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None
    database: Optional[str] = "memgraph"


@dataclass
class TableArtifactsSettings(YAMLSettings):
    # compression: Optional[CompressionSettings] = None
    fmt: Optional[str] = None
    engine: Optional[str] = None
    compression_method: Optional[str] = None
    compression_level: Optional[int] = None


@dataclass
class ArtifactsSettings(YAMLSettings):
    instr_trace: Optional[TableArtifactsSettings] = None
    trace: Optional[TableArtifactsSettings] = None
    table: Optional[TableArtifactsSettings] = None


@dataclass
class OL2Settings(YAMLSettings):
    config_template: Optional[str] = None
    until_step: Optional[str] = None
    target_freq: Optional[int] = None
    target_util: Optional[int] = None


@dataclass
class SynopsysSettings(YAMLSettings):
    pdk: Optional[str] = None
    clock_ns: Optional[float] = None
    core_name: Optional[str] = None


@dataclass
class VivadoSettings(YAMLSettings):
    part: Optional[str] = None
    clock_ns: Optional[float] = None
    core_name: Optional[str] = None


@dataclass
class ASIPSynSettings(YAMLSettings):
    tool: Optional[str] = None
    skip_baseline: Optional[bool] = None
    skip_default: Optional[bool] = None
    skip_shared: Optional[bool] = None
    baseline_use: Optional[str] = None
    ol2: Optional[OL2Settings] = None
    synopsys: Optional[SynopsysSettings] = None


@dataclass
class FPGASynSettings(YAMLSettings):
    tool: Optional[str] = None
    skip_baseline: Optional[bool] = None
    skip_default: Optional[bool] = None
    skip_shared: Optional[bool] = None
    baseline_use: Optional[str] = None
    vivado: Optional[VivadoSettings] = None


@dataclass
class NailgunSettings(YAMLSettings):
    core_name: Optional[str] = None
    ilp_solver: Optional[str] = None
    resource_model: Optional[str] = None
    clock_ns: Optional[float] = None
    schedule_timeout: Optional[float] = None
    refine_timeout: Optional[float] = None
    cell_library: Optional[str] = None
    enable_ol2: Optional[bool] = None
    sched_algo_ms: Optional[bool] = None
    sched_algo_pa: Optional[bool] = None
    sched_algo_mi: Optional[bool] = None
    sched_algo_ra: Optional[bool] = None
    share_resources: Optional[bool] = None  # TODO: partition?
    # label


@dataclass
class HLSSettings(YAMLSettings):
    tool: Optional[str] = None
    nailgun: Optional[NailgunSettings] = None


@dataclass
class StageSettings(YAMLSettings):
    enable: Optional[bool] = None
    defaults: Optional[Dict[str, Any]] = None


@dataclass
class RISCVSettings(YAMLSettings):
    arch: Optional[str] = None
    abi: Optional[str] = None
    xlen: Optional[int] = None  # TODO: auto?


@dataclass
class MlonmcuSettings(YAMLSettings):
    global_isel: Optional[bool] = None
    toolchain: Optional[str] = None
    optimize: Optional[str] = None
    unroll: Optional[bool] = None
    target: Optional[str] = None


@dataclass
class SimSettings(YAMLSettings):
    iss: Optional[str] = None
    iss_perf: Optional[str] = None
    # rtl: Optional[str] = None
    # fpga: Optional[str] = None


@dataclass
class CDFGSettings(YAMLSettings):
    stage: Optional[int] = None
    force_purge_db: Optional[bool] = None


# @dataclass
# class LLVMSettings(YAMLSettings):
#     global_isel: Optional[bool] = None


@dataclass
class ChooseSettings(YAMLSettings):
    check_potential_min_supported: Optional[float] = None
    bb_threshold: Optional[float] = None
    bb_min_weight: Optional[float] = None
    bb_min_supported_weight: Optional[float] = None
    bb_max_num: Optional[int] = None
    bb_min_instrs: Optional[int] = None
    allow_mem: Optional[bool] = None
    allow_loads: Optional[bool] = None
    allow_stores: Optional[bool] = None
    allow_branches: Optional[bool] = None
    allow_compressed: Optional[bool] = None
    allow_custom: Optional[bool] = None
    allow_fp: Optional[bool] = None
    allow_system: Optional[bool] = None
    allow_rvv: Optional[bool] = None


@dataclass
class QuerySettings(YAMLSettings):
    config_file: Optional[str] = None
    limit_results: Optional[int] = None
    # ...


@dataclass
class CoredslSettings(YAMLSettings):
    set_name: Optional[str] = None
    core_name: Optional[str] = None
    splitted: Optional[bool] = None
    base_extensions: Optional[List[str]] = None


@dataclass
class ProgramSettings(YAMLSettings):
    name: Optional[str] = None
    # ...


@dataclass
class ExperimentSettings(YAMLSettings):
    label: Optional[str] = None
    program: Optional[ProgramSettings] = None
    datetime: Optional[str] = None
    # directory: Optional[str] = None
    comment: Optional[str] = None


@dataclass
class ServiceSettings(YAMLSettings):
    hostname: Optional[str] = None
    port: Optional[int] = None
    tags: Optional[dict] = None
    enable: Optional[bool] = None


@dataclass
class DockerSettings(YAMLSettings):
    mlonmcu_image: Optional[str] = None
    etiss_image: Optional[str] = None
    etiss_perf_image: Optional[str] = None
    seal5_image: Optional[str] = None
    hls_image: Optional[str] = None
    enable: Optional[bool] = None


# @dataclass
# class RetargetingSettings(YAMLSettings):
#     iss: Optional[ISSRetargetingSettings] = None
#     llvm: Optional[LLVMRetargetingSettings] = None


@dataclass
class ETISSSettings(YAMLSettings):
    core_name: Optional[str] = None


@dataclass
class DemoSettings(YAMLSettings):
    stages: Optional[Dict[str, StageSettings]] = None
    riscv: Optional[RISCVSettings] = None
    mlonmcu: Optional[MlonmcuSettings] = None
    # llvm: Optional[LLVMSettings] = None
    etiss: Optional[ETISSSettings] = None
    cdfg: Optional[CDFGSettings] = None
    sim: Optional[SimSettings] = None
    choose: Optional[ChooseSettings] = None
    query: Optional[QuerySettings] = None
    coredsl: Optional[CoredslSettings] = None
    hls: Optional[HLSSettings] = None
    asip_syn: Optional[ASIPSynSettings] = None
    fpga_syn: Optional[FPGASynSettings] = None
    experiment: Optional[ExperimentSettings] = None
    docker: Optional[DockerSettings] = None
    service: Optional[ServiceSettings] = None
    # retargeting: Optional[RetargetingSettings] = None


@dataclass
class FlowSettings(YAMLSettings):
    demo: Optional[DemoSettings] = None


@dataclass
class IsaacConfig(YAMLSettings):
    logging: Optional[LoggingSettings] = None
    memgraph: Optional[MemgraphSettings] = None
    flow: Optional[FlowSettings] = None
    artifacts: Optional[ArtifactsSettings] = None



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
from typing import List
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts

logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


def helper(df_operands_full: pd.DataFrame, limit: List[str]):
    df_operands_full["instr"] = df_operands_full["instr"].astype(str)
    df_operands_full["pseudo_instr"] = df_operands_full["instr"]  # start with actual instruction
    df_operands_full.loc[
        (df_operands_full["instr"] == "addi") & (df_operands_full["rs1"] == 0) & (df_operands_full["imm"] != 0),
        "pseudo_instr",
    ] = "li"
    df_operands_full.loc[
        (df_operands_full["instr"] == "addi") & (df_operands_full["imm"] == 0) & (df_operands_full["rs1"] != 0),
        "pseudo_instr",
    ] = "mv"

    df_operands_full.drop(columns=["imm"], inplace=True, errors="ignore")

    df_operands_full["instr"] = df_operands_full["instr"].astype("category")
    df_operands_full["pseudo_instr"] = df_operands_full["pseudo_instr"].astype("category")

    last_written_instr = {}
    rs1_writers = []
    rs2_writers = []

    for _, row in df_operands_full.iterrows():
        rs1 = row["rs1"]
        rs2 = row["rs2"]
        rs1_writers.append(last_written_instr.get(rs1, None) if rs1 is not None else None)
        rs2_writers.append(last_written_instr.get(rs2, None) if rs2 is not None else None)
        rd = row["rd"]
        if rd is not None:
            last_written_instr[rd] = row["pseudo_instr"]  # use pseudo name now

    df_operands_full["rs1_src"] = rs1_writers
    df_operands_full["rs2_src"] = rs2_writers
    df_operands_full.rs1_src.value_counts()

    if len(limit) > 0:
        df_operands_full["rs1_src_flt"] = df_operands_full["rs1_src"].apply(lambda x: x if x in limit else "other")
        df_operands_full["rs2_src_flt"] = df_operands_full["rs2_src"].apply(lambda x: x if x in limit else "other")

    for instr, instr_df in df_operands_full.groupby("pseudo_instr"):
        num = len(instr_df)
        print(">>>", instr, f"[count={num}]", "<<<")
        rs1_src_counts = instr_df.rs1_src_flt.value_counts()
        rs2_src_counts = instr_df.rs2_src_flt.value_counts()
        rs1_src_counts_rel = rs1_src_counts / num
        rs2_src_counts_rel = rs2_src_counts / num
        if len(rs1_src_counts) > 1:
            print("rs1:", rs1_src_counts_rel.to_dict())
        else:
            print("rs1: -")
        if len(rs2_src_counts) > 1:
            print("rs2:", rs2_src_counts_rel.to_dict())
        else:
            print("rs2: -")


def analyze_operands_sources(
    sess: Session,
    force: bool = False,
    limit: List[str] = ["lui", "mv"],
    # filter_instrs: Optional[str] = None,
    # filter_operands: Optional[str] = None,
):
    artifacts = sess.artifacts
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    operands_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "operands")
    assert len(operands_artifacts) == 1
    operands_artifact = operands_artifacts[0]

    assert len(trace_artifact.df) == len(operands_artifact.df)
    operands_df = pd.concat([trace_artifact.df[["instr"]], operands_artifact.df], axis=1).copy()
    del trace_artifact
    del operands_artifact

    helper(operands_df, limit)

    # TODO: gen artifacts
    # operands_hist_artifact = TableArtifact("instr_operands_hist", operands_hist_df, attrs=attrs2)
    # sess.add_artifact(operands_hist_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    analyze_operands_sources(
        sess,
        limit=args.limit.split(","),
        force=args.force,
        # filter_instrs=args.filter_instrs,
        # filter_operands=args.filter_operands,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--limit", type=str, default="li,mv")
    # parser.add_argument("--filter-instrs", type=str, default=None)
    # parser.add_argument("--filter-operands", type=str, default=None)
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import TableArtifact, filter_artifacts, ArtifactFlag
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


# TODO: get names from artifacts?


def check_util(
    sess: Session,
    force: bool = False,
    # names_csv: str = None,
):
    logger.info("Checking ISE utilization...")
    artifacts = sess.artifacts

    # instruction names
    # assert names_csv is not None
    # names_csv = Path(names_csv)
    # assert names_csv.is_file()
    # names_df = pd.read_csv(names_csv)
    # ise_instrs_df = names_df
    ise_instrs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "ise_instrs"
    )
    assert len(ise_instrs_artifacts) == 1
    ise_instrs_artifact = ise_instrs_artifacts[0]
    ise_instrs_df = ise_instrs_artifact.df
    ise_instr_names = ise_instrs_df["instr_lower"].values
    # print("ise_instr_names", ise_instr_names)

    # static counts
    disass_hist_artifacts = filter_artifacts(artifacts, lambda x: x.name == "disass_instrs_hist")
    assert len(disass_hist_artifacts) == 1
    disass_hist_artifact = disass_hist_artifacts[0]
    disass_hist_df = disass_hist_artifact.df
    # print("disass_hist_df", disass_hist_df)
    disass_hist_custom_df = disass_hist_df[disass_hist_df["instr"].apply(lambda x: x.lower() in ise_instr_names)]
    # print("disass_hist_custom_df", disass_hist_custom_df)
    total_disass_instrs = disass_hist_df["count"].sum()

    merged_disass_hist_custom_df = pd.merge(
        ise_instrs_df,
        disass_hist_custom_df,
        how="outer",
        left_on="instr_lower",
        right_on="instr",
        suffixes=("", "_y"),
    )

    # static_count_sum = disass_hist_df["count"].sum()
    # static_count_max = disass_hist_df["count"].max()
    # print("static_count", static_count_sum, static_count_max)
    # static_custom_count_sum = merged_disass_hist_custom_df["count"].sum()
    # static_custom_count_max = merged_disass_hist_custom_df["count"].max()
    # print("static_custom_count", static_custom_count_sum, static_custom_count_max)

    # dynamic counts
    instrs_hist_artifacts = filter_artifacts(artifacts, lambda x: x.name == "instrs_hist")
    assert len(instrs_hist_artifacts) == 1
    instrs_hist_artifact = instrs_hist_artifacts[0]
    instrs_hist_df = instrs_hist_artifact.df
    # print("instrs_hist_df", instrs_hist_df)
    instrs_hist_custom_df = instrs_hist_df[instrs_hist_df["instr"].apply(lambda x: x.lower() in ise_instr_names)]
    # print("instrs_hist_custom_df", instrs_hist_custom_df)
    total_instrs = instrs_hist_df["count"].sum()

    merged_instrs_hist_custom_df = pd.merge(
        ise_instrs_df,
        instrs_hist_custom_df,
        how="outer",
        left_on="instr_lower",
        right_on="instr",
        suffixes=("", "_y"),
    )
    # dynamic_count_sum = instrs_hist_df["count"].sum()
    # dynamic_count_max = instrs_hist_df["count"].max()
    # print("dynamic_instr_count", dynamic_count_sum, dynamic_count_max)
    # dynamic_custom_count_sum = merged_instrs_hist_custom_df["count"].sum()
    # dynamic_custom_count_max = merged_instrs_hist_custom_df["count"].max()
    # print("dynamic_custom_count", dynamic_custom_count_sum, dynamic_custom_count_max)

    # combine
    # TODO: move counts to counts.py?
    merged_disass_hist_custom_df["estimated_reduction"] = merged_disass_hist_custom_df["count"] * (
        merged_disass_hist_custom_df["num_fused_instrs"] - 1
    )
    # TODO: consider instruction size to get memory footprints?
    estimated_total_disass_reduction = merged_disass_hist_custom_df["estimated_reduction"].sum()
    estimated_total_disass_without_ise = total_disass_instrs + estimated_total_disass_reduction
    rel_custom_count = merged_disass_hist_custom_df["rel_count"].sum()
    merged_disass_hist_custom_df["estimated_reduction_rel_scaled"] = (
        merged_disass_hist_custom_df["estimated_reduction"] / estimated_total_disass_reduction
    )
    merged_disass_hist_custom_df["estimated_reduction_rel"] = (
        merged_disass_hist_custom_df["estimated_reduction"] / estimated_total_disass_without_ise
    )
    static_agg_df = pd.DataFrame(
        [
            {
                "instr": None,
                "count": merged_disass_hist_custom_df["count"].sum(),
                "rel_count": merged_disass_hist_custom_df["rel_count"].sum(),
                "estimated_reduction": estimated_total_disass_reduction,
                "estimated_reduction_rel": merged_disass_hist_custom_df["estimated_reduction_rel"].sum(),
                "estimated_reduction_rel_scaled": merged_disass_hist_custom_df["estimated_reduction_rel_scaled"].sum(),
            }
        ]
    )
    static_counts_custom_df = pd.concat(
        [
            static_agg_df,
            merged_disass_hist_custom_df[
                [
                    "instr",
                    "count",
                    "rel_count",
                    "estimated_reduction",
                    "estimated_reduction_rel_scaled",
                    "estimated_reduction_rel",
                ]
            ],
        ]
    )
    merged_disass_hist_custom_df["used"] = merged_disass_hist_custom_df["count"] > 0

    merged_instrs_hist_custom_df["estimated_reduction"] = merged_instrs_hist_custom_df["count"] * (
        merged_instrs_hist_custom_df["num_fused_instrs"] - 1
    )
    estimated_total_reduction = merged_instrs_hist_custom_df["estimated_reduction"].sum()
    estimated_total_instrs_without_ise = total_instrs + estimated_total_reduction
    rel_custom_count = merged_instrs_hist_custom_df["rel_count"].sum()
    merged_instrs_hist_custom_df["estimated_reduction_rel_scaled"] = (
        merged_instrs_hist_custom_df["estimated_reduction"] / estimated_total_reduction
    )
    merged_instrs_hist_custom_df["estimated_reduction_rel"] = (
        merged_instrs_hist_custom_df["estimated_reduction"] / estimated_total_instrs_without_ise
    )
    dynamic_agg_df = pd.DataFrame(
        [
            {
                "instr": None,
                "count": merged_instrs_hist_custom_df["count"].sum(),
                "rel_count": rel_custom_count,
                "estimated_reduction": estimated_total_reduction,
                "estimated_reduction_rel": merged_instrs_hist_custom_df["estimated_reduction_rel"].sum(),
                "estimated_reduction_rel_scaled": merged_instrs_hist_custom_df["estimated_reduction_rel_scaled"].sum(),
            }
        ]
    )
    dynamic_counts_custom_df = pd.concat(
        [
            dynamic_agg_df,
            merged_instrs_hist_custom_df[
                [
                    "instr",
                    "count",
                    "rel_count",
                    "estimated_reduction",
                    "estimated_reduction_rel_scaled",
                    "estimated_reduction_rel",
                ]
            ],
        ]
    )
    merged_instrs_hist_custom_df["used"] = merged_instrs_hist_custom_df["count"] > 0

    ise_util_df = pd.merge(
        merged_disass_hist_custom_df,
        merged_instrs_hist_custom_df,
        on="instr",
        suffixes=("_static", "_dynamic"),
    )
    ise_util_df["used_only_static"] = ise_util_df["used_static"] & ~ise_util_df["used_dynamic"]
    n_instrs = len(ise_instrs_df)
    n_used_static = ise_util_df["used_static"].sum()
    n_used_dynamic = ise_util_df["used_dynamic"].sum()
    n_used_static_rel = n_used_static / n_instrs
    n_used_dynamic_rel = n_used_dynamic / n_instrs
    ise_util_agg_df = pd.DataFrame(
        [
            {
                "instr": None,
                "n_total": n_instrs,
                "n_used_static": n_used_static,
                "n_used_dynamic": n_used_dynamic,
                "n_used_static_rel": n_used_static_rel,
                "n_used_dynamic_rel": n_used_dynamic_rel,
            }
        ]
    )
    ise_util_df = pd.concat(
        [
            ise_util_agg_df,
            ise_util_df[["instr", "used_static", "used_dynamic", "used_only_static"]],
        ]
    )

    attrs = {}
    static_counts_custom_artifact = TableArtifact("static_counts_custom", static_counts_custom_df, attrs=attrs)
    dynamic_counts_custom_artifact = TableArtifact("dynamic_counts_custom", dynamic_counts_custom_df, attrs=attrs)
    ise_util_artifact = TableArtifact("ise_util", ise_util_df, attrs=attrs)

    sess.add_artifact(static_counts_custom_artifact, override=force)
    sess.add_artifact(dynamic_counts_custom_artifact, override=force)
    sess.add_artifact(ise_util_artifact, override=force)

    # TODO
    # pc2bb_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "pc2bb")
    # assert len(pc2bb_artifacts) == 1
    # pc2bb_artifact = pc2bb_artifacts[0]
    # pc2bb_df = pc2bb_artifact.df

    # symbol_map_artifacts = filter_artifacts(
    #     artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "symbol_map"
    # )
    # symbol_map_df = None
    # if len(symbol_map_artifacts) > 0:
    #     assert len(symbol_map_artifacts) == 1
    #     symbol_map_artifact = symbol_map_artifacts[0]
    #     symbol_map_df = symbol_map_artifact.df

    # llvm_bbs_artifacts = filter_artifacts(
    #     artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs_new"
    # )
    # llvm_bbs_df = None
    # if len(llvm_bbs_artifacts) > 0:
    #     assert len(llvm_bbs_artifacts) == 1
    #     llvm_bbs_artifact = llvm_bbs_artifacts[0]
    #     llvm_bbs_df = llvm_bbs_artifact.df.copy()

    # plots_dir = sess.directory / "plots"
    # plots_dir.mkdir(exist_ok=True)
    # # TODO: use threshold
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    # print("elf_artifacts", elf_artifacts)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    trace_df = trace_artifact.df.copy()
    trace_len = len(trace_df)

    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "func2pc")
    assert len(func2pc_artifacts) == 1
    func2pc_artifact = func2pc_artifacts[0]
    func2pc_df = func2pc_artifact.df.copy()
    func2pc_df[["start", "end"]] = func2pc_df["pc_range"].apply(pd.Series)

    def helper(pc):
        # print("x", x)
        # print("!", func2pc_df["start"] >= x[0] & func2pc_df["end"] <= x[1])
        matches = func2pc_df[func2pc_df["start"] <= pc]
        matches = matches[matches["end"] >= pc]
        # print("matches", matches)
        if len(matches) == 0:
            return None
        # print("len(matches)", len(matches))
        # assert len(matches) == 1
        # match_ = matches.iloc[0]
        # func = match_["func"]
        # TODO: assert that alias!
        funcs = set(matches["func"].values)
        FIRST_ONLY = True
        if FIRST_ONLY:
            if len(funcs) > 0:
                funcs = list(funcs)[0]
            else:
                funcs = None
        # return func
        return funcs

    def helper2(pc):
        # print("x", x)
        # print("!", func2pc_df["start"] >= x[0] & func2pc_df["end"] <= x[1])
        matches = trace_df[trace_df["pc"] == pc]
        # print("matches", matches)
        if len(matches) == 0:
            return None
        # print("len(matches)", len(matches))
        # assert len(matches) == 1
        # match_ = matches.iloc[0]
        # func = match_["func"]
        # TODO: assert that alias!
        instrs = set(matches["instr"].values)
        # print("instrs", instrs)
        assert len(instrs) == 1
        instr = list(instrs)[0]
        return instr

    pc_df = trace_df["pc"].value_counts().to_frame("count").reset_index()
    # trace_df["func_name"] = trace_df["pc"].apply(helper)
    pc_df["instr"] = pc_df["pc"].apply(helper2)
    pc_df["func_name"] = pc_df["pc"].apply(helper)
    CUSTOM_ONLY = True
    if CUSTOM_ONLY:
        pc_df = pc_df[pc_df["instr"].apply(lambda x: x.lower() in ise_instr_names)]

    # instrs_per_func = pc_df.groupby("func_name")["instr"].value_counts().to_dict()
    # print("instrs_per_func", instrs_per_func)
    func_ise_counts = {}
    for func_name, group_df in pc_df.groupby("func_name"):
        counts = group_df.groupby("instr")["count"].sum()
        # counts = group_df["instr"].value_counts().to_dict()
        # print("counts", counts)
        func_ise_counts[func_name] = counts.to_dict()
    # print("func_ise_counts", func_ise_counts)
    func_ise_rel_counts = {
        func_name: {instr: count / trace_len for instr, count in counts.items()}
        for func_name, counts in func_ise_counts.items()
    }
    # print("func_ise_rel_counts", func_ise_rel_counts)
    dynamic_counts_custom_per_func_df = pd.DataFrame(
        [{"func_name": func_name, **counts} for func_name, counts in func_ise_counts.items()]
    ).fillna(0)
    # print("dynamic_counts_custom_per_func_df", dynamic_counts_custom_per_func_df)
    dynamic_rel_counts_custom_per_func_df = pd.DataFrame(
        [{"func_name": func_name, **counts} for func_name, counts in func_ise_rel_counts.items()]
    ).fillna(0)
    # print("dynamic_rel_counts_custom_per_func_df", dynamic_rel_counts_custom_per_func_df)

    dynamic_counts_custom_per_func_artifact = TableArtifact(
        "dynamic_counts_custom_per_func", dynamic_counts_custom_per_func_df, attrs=attrs
    )
    dynamic_rel_counts_custom_per_func_artifact = TableArtifact(
        "dynamic_rel_counts_custom_per_func",
        dynamic_rel_counts_custom_per_func_df,
        attrs=attrs,
    )

    sess.add_artifact(dynamic_counts_custom_per_func_artifact, override=force)
    sess.add_artifact(dynamic_rel_counts_custom_per_func_artifact, override=force)

    # static per_func
    disass_table_artifacts = filter_artifacts(artifacts, lambda x: x.name == "disass")
    assert len(disass_table_artifacts) == 1
    disass_df = disass_table_artifacts[0].df.copy()
    disass_len = len(disass_df)
    CUSTOM_ONLY = True
    if CUSTOM_ONLY:
        disass_df = disass_df[disass_df["instr"].apply(lambda x: x.lower() in ise_instr_names)]

    static_pc_df = disass_df
    static_pc_df["func_name"] = static_pc_df["pc"].apply(helper)
    # print("static_pc_df", static_pc_df)

    static_func_ise_counts = {}
    for func_name, group_df in static_pc_df.groupby("func_name"):
        # print("func_name", func_name)
        counts = group_df["instr"].value_counts()
        # counts = group_df["instr"].value_counts().to_dict()
        # print("counts", counts)
        static_func_ise_counts[func_name] = counts.to_dict()
    # print("static_func_ise_counts", static_func_ise_counts)
    static_func_ise_rel_counts = {
        func_name: {instr: count / disass_len for instr, count in counts.items()}
        for func_name, counts in static_func_ise_counts.items()
    }
    # print("static_func_ise_rel_counts", static_func_ise_rel_counts)
    static_counts_custom_per_func_df = pd.DataFrame(
        [{"func_name": func_name, **counts} for func_name, counts in static_func_ise_counts.items()]
    ).fillna(0)
    # print("static_counts_custom_per_func_df", static_counts_custom_per_func_df)
    static_rel_counts_custom_per_func_df = pd.DataFrame(
        [{"func_name": func_name, **counts} for func_name, counts in static_func_ise_rel_counts.items()]
    ).fillna(0)
    # print("static_rel_counts_custom_per_func_df", static_rel_counts_custom_per_func_df)

    static_counts_custom_per_func_artifact = TableArtifact(
        "static_counts_custom_per_func", static_counts_custom_per_func_df, attrs=attrs
    )
    static_rel_counts_custom_per_func_artifact = TableArtifact(
        "static_rel_counts_custom_per_func",
        static_rel_counts_custom_per_func_df,
        attrs=attrs,
    )

    sess.add_artifact(static_counts_custom_per_func_artifact, override=force)
    sess.add_artifact(static_rel_counts_custom_per_func_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    check_util(
        sess,
        force=args.force,
        # names_csv=args.names_csv,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # parser.add_argument("--names-csv", required=True)
    # TODO: !
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.generate.ise.check_ise_potential_per_llvm_bb import (
    check_ise_potential_per_llvm_bb,
)
from isaac_toolkit.generate.ise.check_ise_potential import check_ise_potential
from isaac_toolkit.generate.ise.choose_bbs import choose_bbs

logger = get_logger()


def analyze_bottleneck_bbs(sess: Session, force: bool = False, progress: bool = False):
    logger.info("Analyzing bottleneck BBs...")
    config = sess.config
    flow_config = config.flow
    assert flow_config is not None
    demo_config = flow_config.demo
    assert demo_config is not None
    choose_config = demo_config.choose
    assert choose_config is not None
    check_ise_potential(
        sess,
        min_supported=choose_config.check_potential_min_supported,
        allow_mem=choose_config.allow_mem,
        allow_loads=choose_config.allow_loads,
        allow_stores=choose_config.allow_stores,
        allow_branches=choose_config.allow_branches,
        allow_compressed=choose_config.allow_compressed,
        allow_custom=choose_config.allow_custom,
        allow_fp=choose_config.allow_fp,
        allow_system=choose_config.allow_system,
        force=force,
    )
    check_ise_potential_per_llvm_bb(
        sess,
        min_supported=choose_config.check_potential_min_supported,
        allow_mem=choose_config.allow_mem,
        allow_loads=choose_config.allow_loads,
        allow_stores=choose_config.allow_stores,
        allow_branches=choose_config.allow_branches,
        allow_compressed=choose_config.allow_compressed,
        allow_custom=choose_config.allow_custom,
        allow_fp=choose_config.allow_fp,
        allow_system=choose_config.allow_system,
        force=force,
    )
    choose_bbs(
        sess,
        threshold=choose_config.bb_threshold,
        min_weight=choose_config.bb_min_weight,
        min_supported_weight=choose_config.bb_min_supported_weight,
        min_instrs=choose_config.bb_min_instrs,
        max_num=choose_config.bb_max_num,
        force=force,
    )


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    analyze_bottleneck_bbs(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import sys
import argparse
from typing import Optional
from math import ceil, log2
from pathlib import Path
from collections import defaultdict

import pandas as pd
import matplotlib.pyplot as plt

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def analyze_instr_operands(
    sess: Session,
    force: bool = False,
    imm_only: bool = True,
    signed: bool = False,
    subplots: bool = False,
    plot_fmt: str = "pdf",
    topk: int = 20,
    filter_instrs: Optional[str] = None,
    filter_operands: Optional[str] = None,
):
    logger.info("Analyzing instruction operands...")
    artifacts = sess.artifacts
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    operands_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "operands")
    assert len(operands_artifacts) == 1
    operands_artifact = operands_artifacts[0]
    # filter_instrs = "addi"
    # filter_operands = "imm"

    assert len(trace_artifact.df) == len(operands_artifact.df)
    operands_df = pd.concat([trace_artifact.df[["instr"]], operands_artifact.df], axis=1).copy()

    attrs2 = {
        "trace": trace_artifact.name,
        "kind": "histogram",
        "by": __name__,
    }

    del trace_artifact

    operand_names = sorted([x for x in operands_df.columns if x != "instr"])

    to_drop = {"rd", "rs1", "rs2", "rs3"} if imm_only else set()
    if filter_operands is not None:
        keep = set(filter(lambda x: re.compile(filter_operands).match(x), operand_names))
        drop = set(operand_names) - keep
        to_drop |= drop
        operand_names = sorted(list(keep))
    if to_drop:
        for col in to_drop:
            if col in operands_df.columns:
                operands_df.drop(columns=[col], inplace=True)
    instrs = sorted(operands_df["instr"].unique())

    to_keep = instrs
    if filter_instrs is not None:
        to_keep = list(filter(lambda x: re.compile(filter_instrs).match(x), instrs))

    operand_names = sorted([x for x in operands_df.columns if x != "instr"])
    to_keep = [
        instr_name
        for instr_name in to_keep
        if not pd.isna(operands_df[operands_df["instr"] == instr_name][operand_names]).all().all()
    ]

    if len(to_keep) < len(instrs):
        res = operands_df[operands_df["instr"].astype(str).isin(to_keep)]
        operands_df = res

    instrs = sorted(operands_df["instr"].unique())

    # operands_artifact = TableArtifact("instr_operands", operands_df, attrs=attrs)
    # sess.add_artifact(operands_artifact, override=force)

    plot = True
    if plot:
        plots_dir = sess.directory / "plots"
        plots_dir.mkdir(exist_ok=True)

    operands_hist_data = []
    # TODO: share code with below
    # TODO: split artifact creation and plotting
    plot_data = defaultdict(dict)
    for i, instr_name in enumerate(instrs):
        instr_df = operands_df[operands_df["instr"] == instr_name]
        for j, operand_name in enumerate(operand_names):
            if operand_name not in instr_df.columns or pd.isna(instr_df[operand_name]).all():
                continue
            counts = instr_df[operand_name].value_counts()
            bit_counts = instr_df[operand_name].apply(lambda x: max(1, ceil(log2(1 + x)))).value_counts()
            plot_data[operand_name][instr_name] = (counts, bit_counts)
            operands_hist_data.append(
                {
                    "instr": instr_name,
                    "op": operand_name,
                    "counts": counts.to_dict(),
                    "bit_counts": bit_counts.to_dict(),
                }
            )

    if plot:
        num_operands = len(operand_names)
        assert num_operands > 0
        if subplots:
            num_instrs = len(instrs)
            assert num_instrs > 0
            # VALUES
            fig, axes = plt.subplots(
                nrows=num_operands,
                ncols=num_instrs,
                figsize=(3 * num_instrs, 2 * num_operands),
            )
            plt.tight_layout()
            if num_instrs == 1 and num_operands == 1:
                axes = [[axes]]
            elif num_instrs == 1 or num_operands == 1:
                axes = [axes]
            for i, operand_name in enumerate(operand_names):
                for j, instr_name in enumerate(instrs):
                    axes[i][j].axes.xaxis.set_ticklabels([])
                    axes[i][j].axes.yaxis.set_ticklabels([])
                    axes[i][j].tick_params(axis="x", labelsize=8)
                    axes[i][j].tick_params(axis="y", labelsize=8)
            for operand_name, plot_data_ in plot_data.items():
                i = operand_names.index(operand_name)
                for instr_name, plot_data__ in plot_data_.items():
                    j = instrs.index(instr_name)
                    val_counts = plot_data__[0]
                    val_counts.head(n=topk).plot(ax=axes[i][j], kind="bar")
                    axes[i][j].set_xlabel("", fontsize=8)
            for i, operand_name in enumerate(operand_names):
                axes[i][0].set_ylabel(operand_name, fontsize=8)
            for j, instr_name in enumerate(instrs):
                axes[0][j].set_title(instr_name)
                axes[-1][j].set_xlabel("Value", fontsize=8)
            plot_file = plots_dir / f"operands-hist.{plot_fmt}"
            fig.savefig(plot_file, bbox_inches="tight")
            plt.close()
            # BITS
            fig, axes = plt.subplots(
                nrows=num_operands,
                ncols=num_instrs,
                figsize=(3 * num_instrs, 2 * num_operands),
            )
            plt.tight_layout()
            if num_instrs == 1 and num_operands == 1:
                axes = [[axes]]
            elif num_instrs == 1 or num_operands == 1:
                axes = [axes]
            for i, operand_name in enumerate(operand_names):
                for j, instr_name in enumerate(instrs):
                    axes[i][j].axes.xaxis.set_ticklabels([])
                    axes[i][j].axes.yaxis.set_ticklabels([])
                    axes[i][j].tick_params(axis="x", labelsize=8)
                    axes[i][j].tick_params(axis="y", labelsize=8)
            for operand_name, plot_data_ in plot_data.items():
                i = operand_names.index(operand_name)
                for instr_name, plot_data__ in plot_data_.items():
                    j = instrs.index(instr_name)
                    bit_counts = plot_data__[1]
                    bit_counts.head(n=topk).plot(ax=axes[i][j], kind="bar")
                    axes[i][j].set_xlabel("", fontsize=8)
            for i, operand_name in enumerate(operand_names):
                axes[i][0].set_ylabel(operand_name, fontsize=8)
            for j, instr_name in enumerate(instrs):
                axes[0][j].set_title(instr_name)
                axes[-1][j].set_xlabel("#Bits", fontsize=8)
            plot_file = plots_dir / f"operands-hist-bits.{plot_fmt}"
            fig.savefig(plot_file, bbox_inches="tight")
            plt.close()
        else:
            plot_data2 = {
                instr_name: {
                    operand_name: plot_data[operand_name][instr_name]
                    for operand_name in operand_names
                    if operand_name in plot_data and instr_name in plot_data[operand_name].keys()
                }
                for instr_name in instrs
            }
            for instr_name, plot_data_ in plot_data2.items():
                operand_names_ = sorted(list(plot_data_.keys()))
                num_operands_ = len(operand_names_)
                if num_operands_ == 0:
                    continue
                # VALUES
                fig, axes = plt.subplots(nrows=num_operands_, ncols=1, figsize=(10, 2 * num_operands))
                plt.tight_layout()
                plot_data = {}
                if num_operands_ == 1:
                    axes = [axes]
                for operand_name, plot_data__ in plot_data_.items():
                    i = operand_names_.index(operand_name)
                    val_counts = plot_data__[0]
                    val_counts.head(n=topk).plot(ax=axes[i], kind="bar")
                    axes[i].tick_params(axis="x", labelsize=8)
                    axes[i].tick_params(axis="y", labelsize=8)
                    axes[i].set_ylabel(operand_name, fontsize=8)
                    axes[i].set_xlabel("", fontsize=8)
                axes[0].set_title(instr_name)
                axes[-1].set_xlabel("Value", fontsize=8)
                plot_file = plots_dir / f"operands-hist-{instr_name}.{plot_fmt}"
                fig.savefig(plot_file, bbox_inches="tight")
                plt.close()
                # BITS
                fig, axes = plt.subplots(nrows=num_operands_, ncols=1, figsize=(10, 2 * num_operands))
                plt.tight_layout()
                plot_data = {}
                if num_operands_ == 1:
                    axes = [axes]
                for operand_name, plot_data__ in plot_data_.items():
                    i = operand_names_.index(operand_name)
                    bit_counts = plot_data__[1]
                    bit_counts.head(n=topk).plot(ax=axes[i], kind="bar")
                    axes[i].tick_params(axis="x", labelsize=8)
                    axes[i].tick_params(axis="y", labelsize=8)
                    axes[i].set_ylabel(operand_name, fontsize=8)
                    axes[i].set_xlabel("", fontsize=8)
                axes[0].set_title(instr_name)
                axes[-1].set_xlabel("#Bits", fontsize=8)
                plot_file = plots_dir / f"operands-hist-{instr_name}-bits.{plot_fmt}"
                fig.savefig(plot_file, bbox_inches="tight")
                plt.close()
    operands_hist_df = pd.DataFrame(operands_hist_data)

    operands_hist_artifact = TableArtifact("instr_operands_hist", operands_hist_df, attrs=attrs2)
    sess.add_artifact(operands_hist_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    analyze_instr_operands(
        sess,
        force=args.force,
        signed=args.signed,
        imm_only=args.imm_only,
        plot_fmt=args.fmt,
        topk=args.topk,
        filter_instrs=args.filter_instrs,
        filter_operands=args.filter_operands,
        subplots=args.subplots,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--signed", action="store_true")
    parser.add_argument("--imm-only", action="store_true")
    parser.add_argument("--subplots", action="store_true")
    parser.add_argument("--fmt", type=str, default="pdf")
    parser.add_argument("--topk", type=int, default=20)
    parser.add_argument("--filter-instrs", type=str, default=None)
    parser.add_argument("--filter-operands", type=str, default=None)
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import networkx as nx

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import GraphArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def maxmiso_algo(G):
    logger.info("Running MaxMISO algorithm...")
    # print("algo")
    # print("G", G, dir(G))
    G = G.copy()
    if G.number_of_nodes() == 0:
        return []
    mapping = dict(zip(G.nodes.keys(), range(len(G.nodes))))
    G = nx.relabel_nodes(G, mapping)
    topo = list(reversed(list(nx.topological_sort(G))))
    # print("topo", topo)
    invalid = [False] * len(topo)
    fanout = [0] * len(topo)
    processed = [False] * len(topo)
    fanout_org = [0] * len(topo)
    for node in G.nodes:
        # print("node", node)
        # print("node_data", G.nodes[node])
        op_type = G.nodes[node]["properties"].get("op_type", "ot")
        # print("op_type", op_type)
        od = G.out_degree(node)
        # is_output = od == 0
        is_output = op_type == "output"
        if is_output:
            od += 1
        # print("od", od)
        fanout_org[topo.index(node)] = od
        # is_input = G.nodes[node].get("label") == "Const"  # TODO: add to db
        is_input = op_type == "input"
        is_const = op_type == "constant"
        is_label = op_type == "label"
        load_instrs = ["LB", "LBU", "LH", "LHU", "LW"]  # TODO: do not hardcode
        is_load = G.nodes[node].get("label") in load_instrs  # TODO: add to db
        is_invalid = is_input or is_load or is_const or is_label
        # is_invalid = False
        if is_invalid:
            invalid[topo.index(node)] = True
    max_misos = []
    for node in topo:
        # print("node", node)
        if processed[topo.index(node)]:
            continue
        fanout = fanout_org.copy()
        processed[topo.index(node)] = True
        if invalid[topo.index(node)]:
            continue
        max_miso = [False] * len(topo)
        max_miso[topo.index(node)] = True

        def generate_max_miso(node, count):
            ins = G.in_edges(node)
            # print("ins", ins)
            for src, dest in ins:
                # print("src", src)
                # print("dest", dest)
                fanout[topo.index(src)] -= 1
                if not invalid[topo.index(src)] and fanout[topo.index(src)] == 0:
                    max_miso[topo.index(src)] = True
                    processed[topo.index(src)] = True
                    count = generate_max_miso(src, count + 1)
            return count

        size = generate_max_miso(node, 1)
        # size = generate_max_miso(node, G, max_miso, 1, invalid, fanout, processed)
        # print("size", size)
        if size > 1:

            def calc_inputs(max_miso):
                # print("calc_inputs", max_miso)
                inputs = [False] * len(topo)
                ret = 0
                max_miso_nodes = [topo[i] for i, val in enumerate(max_miso) if val]
                # print("mmn", max_miso_nodes)
                for node in max_miso_nodes:
                    # print("node", node)
                    ins = G.in_edges(node)
                    # print("ins", ins)
                    for in_ in ins:
                        # print("in_", in_, G.nodes[in_[0]].get("label"))
                        src = in_[0]
                        if not max_miso[topo.index(src)] and not inputs[topo.index(src)]:
                            ret += 1
                            inputs[topo.index(src)] = True
                # print("ret", ret)
                # input("1")
                return ret

            def calc_outputs(max_miso):
                # print("calc_outputs", max_miso)
                # outputs = [False] * len(topo)
                ret = 0
                max_miso_nodes = [topo[i] for i, val in enumerate(max_miso) if val]
                # print("mmn", max_miso_nodes)
                for node in max_miso_nodes:
                    # print("node", node)
                    if G.nodes[node]["properties"]["op_type"] == "output":
                        # print("A")
                        ret += 1
                    else:
                        # print("B")
                        outs = G.out_edges(node)
                        # print("outs", outs)
                        for out_ in outs:
                            # print("out_", out_, G.nodes[out_[0]].get("label"))
                            dst = out_[1]
                            if not max_miso[topo.index(dst)]:
                                ret += 1
                # print("ret", ret)
                # input("1")
                return ret

            # num_inputs = calc_inputs(max_miso)
            # print("num_inputs", num_inputs)
            # num_outputs = calc_outputs(max_miso)
            # print("num_outputs", num_outputs)
            # input(">")
            max_misos.append(max_miso)
    # print("max_misos", max_misos, len(max_misos))
    from itertools import compress

    max_misos_ = [list(compress(topo, max_miso)) for max_miso in max_misos]
    # print("max_misos_", max_misos_)
    # max_misos__ = [nx.subgraph_view(G, filter_node=lambda node: max_miso[topo.index(node)]) for max_miso in max_misos]
    # max_misos__ = [nx.subgraph_view(G, filter_node=lambda node: node in max_miso) for max_miso in max_misos_]
    # max_misos__ = [nx.subgraph_view(G, filter_node=lambda node: node in max_miso) for max_miso in max_misos_]
    # print("G", type(G), G)
    # print("G.", type(G.subgraph(max_misos_[0])), G.subgraph(max_misos_[0])))
    max_misos__ = [G.subgraph(max_miso) for max_miso in max_misos_]
    reverse_mapping = {v: k for k, v in mapping.items()}
    G = nx.relabel_nodes(G, reverse_mapping)
    max_misos__ = [nx.relabel_nodes(max_miso, reverse_mapping) for max_miso in max_misos__]
    # print("max_misos__", max_misos__)
    # for i, mig in enumerate(max_misos__):
    #     print("i,mig", i, mig)
    #     write_dot(mig, f"maxmiso{i}.dot")
    #     labeldict = {node: mig.nodes[node]["label"] for node in mig.nodes}
    #     print("labeldict", labeldict)
    #     nx.draw(mig, labels=labeldict, with_labels=True)
    #     plt.savefig(f"maxmiso{i}.png")
    #     plt.close()
    labeldict = {node: G.nodes[node]["label"] for node in G.nodes}
    # print("labeldict", labeldict)
    # nx.draw(G, labels=labeldict, with_labels=True)
    # plt.savefig(f"full.png")
    # plt.close()

    # print("invalid", invalid)
    # print("fanout", fanout)
    # print("processed", processed)
    # print("fanout_org", fanout_org)
    return max_misos__


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    override = args.force
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    graphs = sess.graphs
    # print("graphs", graphs)
    # cdfg = filter_artifacts(graphs, lambda x: x.name == args.graph_name)
    dfgs = filter_artifacts(graphs, lambda x: x.attrs.get("kind") == "dfg")
    # print("cdfg", dfgs)
    assert len(dfgs) > 0, "No DFGs found!"
    for dfg in dfgs:
        module_name = dfg.attrs["module_name"]
        func_name = dfg.attrs["func_name"]
        bb_name = dfg.attrs["bb_name"]
        graph = dfg.graph
        # print("graph", graph)

        def filter_graph(G, func_name, bb_name):
            # TODO: module_name
            view = nx.subgraph_view(
                G,
                filter_node=lambda node: (bb_name is None or G.nodes[node]["properties"].get("basic_block") == bb_name)
                and (func_name is None or G.nodes[node]["properties"].get("func_name") == func_name)
                and "%bb" not in G.nodes[node].get("label"),
            )
            G_ = G.subgraph([node for node in view.nodes])
            # G__ = nx.subgraph_view(G_, filter_edge=lambda n1, n2: G_[n1][n2]["type"] == "DFG")
            return G_

        graph = filter_graph(graph, func_name=func_name, bb_name=bb_name)
        result = maxmiso_algo(graph)
        # print("result", result)
        for i, maxmiso in enumerate(result):
            attrs = {
                "kind": "maxmiso",
                "mod_name": module_name,
                "func_name": func_name,
                "bb_name": bb_name,
                "by": "isaac_toolkit.algorithm.ise.identification.maxmiso",
            }
            artifact = GraphArtifact(
                f"{module_name}/{func_name}/{bb_name}/maxmiso/{i}",
                nx.DiGraph(maxmiso),
                attrs=attrs,
            )
            # print("artifact", artifact)
            sess.add_artifact(artifact, override=override)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    # parser.add_argument("graph_name", default="memgraph_mir_cdfg")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
from pathlib import Path

# from collections import defaultdict

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


def collect_pcs(trace_df):
    pcs_df = (
        trace_df["pc"]
        .value_counts()
        .rename_axis("pc")  # makes 'pc' a column
        .reset_index(name="count")  # turns counts into a column
    )
    total_count = pcs_df["count"].sum()
    pcs_df["rel_count"] = pcs_df["count"] / total_count
    pcs_df.sort_values("count", ascending=False, inplace=True)
    return pcs_df


def create_pc_hist(sess: Session, force: bool = False):
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    # print("elf_artifacts", elf_artifacts)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]

    pcs_df = collect_pcs(trace_artifact.df)

    attrs = {
        "trace": trace_artifact.name,
        "kind": "histogram",
        "by": __name__,
    }

    pcs_artifact = TableArtifact("pcs_hist", pcs_df, attrs=attrs)
    sess.add_artifact(pcs_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    create_pc_hist(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def analyze_linker_map_helper(mapFile):
    ret = []
    data = mapFile.toJson(humanReadable=False)
    segments = data["segments"]
    for segment in segments:
        segment_name = segment["name"]
        files = segment["files"]
        for file in files:
            filepath = file["filepath"]
            if "(" in filepath:  # TODO: use regex instead
                library, obj = filepath[:-1].split("(", 1)
            else:
                library = None
                obj = filepath
            obj_short = Path(obj).name
            library_short = Path(library).name if library is not None else library
            section_type = file["sectionType"]
            symbols = file["symbols"]
            for symbol in symbols:
                symbol_name = symbol["name"]
                new = {
                    "segment": segment_name,
                    "section": section_type,
                    "symbol": symbol_name,
                    "library": library_short,
                    "library_full": library,
                    "object": obj_short,
                    "object_full": obj,
                }
                ret.append(new)
    return ret


def analyze_linker_map(sess: Session, force: bool = False):
    logger.info("Analyzing linker map...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    linker_map_artifacts = filter_artifacts(
        artifacts, lambda x: x.name.endswith(".map") and x.flags & ArtifactFlag.PYTHON
    )
    assert len(linker_map_artifacts) == 1
    linker_map_artifact = linker_map_artifacts[0]
    mapFile = linker_map_artifact.data

    symbol_map = analyze_linker_map_helper(mapFile)
    symbol_map_df = pd.DataFrame(
        symbol_map,
        columns=[
            "segment",
            "section",
            "symbol",
            "object",
            "object_full",
            "library",
            "library_full",
        ],
    )

    attrs = {
        "kind": "mapping",
        "by": __name__,
    }

    symbol_map_artifact = TableArtifact("symbol_map", symbol_map_df, attrs=attrs)
    sess.add_artifact(symbol_map_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    analyze_linker_map(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import time
import tempfile
import argparse
from pathlib import Path

from isaac_toolkit.session import Session
from isaac_toolkit.session.session import get_table_artifact_ext
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts
from isaac_toolkit.session.config import ArtifactsSettings
from isaac_toolkit.report.sess_mem_usage import get_artifact_size
from isaac_toolkit.report.sess_disk_usage import get_file_size


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    # print("trace_artifact", trace_artifact)
    t0 = time.time()
    trace_df = trace_artifact.df
    t1 = time.time()
    del trace_df
    artifact_size = get_artifact_size(trace_artifact)
    print("artifact_size", round(artifact_size / 1000, 2))
    # print("trace_df", trace_df)
    artifacts_settings_dict = {
        "instr_trace": {
            "fmt": args.fmt,
            "engine": args.engine,
            "compression_method": args.method,
            "compression_level": int(args.level) if args.level is not None else None,
        }
    }
    artifacts_settings = ArtifactsSettings.from_dict(artifacts_settings_dict)
    print("artifacts_settings", artifacts_settings)
    with tempfile.TemporaryDirectory() as tmpdirname:
        print("created temporary directory", tmpdirname)
        ext = get_table_artifact_ext(artifacts_settings.instr_trace)
        fname = f"{trace_artifact.name}.{ext}"
        temp_file = Path(tmpdirname) / fname
        t2 = time.time()
        trace_artifact._save(temp_file, artifacts_settings=artifacts_settings)
        # input("!")
        t3 = time.time()
        compressed_size = get_file_size(temp_file)
        print("compressed_size", round(compressed_size / 1000, 2))
        t4 = time.time()
        trace_artifact._load(temp_file)
        t5 = time.time()
    # ts = [t0, t1, t2, t3, t4, t5]
    # print("ts", ts)
    tds = {"t_cache": round(t1 - t0, 3), "t_save": round(t3 - t2, 3), "t_load": round(t5 - t4, 3)}
    print("tds", tds)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--fmt", default="pickle")
    parser.add_argument("--engine", default=None)
    parser.add_argument("--method", default=None)
    parser.add_argument("--level", default=None)
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import argparse
import posixpath
from pathlib import Path
from collections import defaultdict

import pandas as pd
from elftools.elf.elffile import ELFFile

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.utils.demangle import unmangle_helper

logger = get_logger()


def parse_dwarf(elf_path):
    # func2pcs_data = []
    func_ranges = {}
    # the mapping between source file and its function
    srcFile_func_dict = defaultdict(lambda: [set(), set(), set()])
    # the mapping between program counter and source line
    pc_to_source_line_mapping = defaultdict(list)
    with open(elf_path, "rb") as f:
        elffile = ELFFile(f)

        # mapping function symbol to pc range
        symtab = elffile.get_section_by_name(".symtab")
        if symtab is None:
            raise RuntimeError("No symbol table found (.symtab missing)")

        funcs_by_section = defaultdict(list)  # shndx -> list of (start, size, name, symbol)
        for symbol in symtab.iter_symbols():
            try:
                st_type = symbol["st_info"]["type"]
            except Exception:
                continue
            if st_type != "STT_FUNC":
                continue
            start = symbol["st_value"]
            size = symbol["st_size"]
            shndx = symbol["st_shndx"]
            name = symbol.name or "<anon>"
            funcs_by_section[shndx].append((start, size, name, symbol))

        # 3) For each section, sort by start addr and compute ranges
        for shndx, syms in funcs_by_section.items():
            # skip special indices (we still handle them but there might be no section)
            section = None
            try:
                if isinstance(shndx, int):
                    section = elffile.get_section(shndx)
                else:
                    # PyElfTools may give SHN_UNDEF as string, leave section None
                    section = None
            except Exception:
                section = None

            syms_sorted = sorted(syms, key=lambda x: x[0])
            for idx, (start, size, name, symbol) in enumerate(syms_sorted):

                if size and size != 0:
                    end = start + size - 1
                    func_ranges[name] = (start, end)
                    continue

                # size == 0: try to infer from next symbol in same section
                end = None
                # find next symbol with start > this start
                next_start = None
                for j in range(idx + 1, len(syms_sorted)):
                    candidate_start = syms_sorted[j][0]
                    if candidate_start > start:
                        next_start = candidate_start
                        break
                if next_start is not None:
                    end = next_start - 1
                elif section is not None:
                    # fallback to section end
                    sec_start = section["sh_addr"]
                    sec_size = section["sh_size"]
                    end = sec_start + sec_size - 1
                else:
                    # last resort: set end == start (or None)
                    end = start

                func_ranges[name] = (start, end)

        func2pcs_data = list(func_ranges.items())
        # mapping source file to function
        if not elffile.has_dwarf_info():
            logger.error("ELF file has no DWARF info!")
            return func2pcs_data, None, None

        dwarfinfo = elffile.get_dwarf_info()

        def lpe_filename(line_program, file_index):
            # Retrieving the filename associated with a line program entry
            # involves two levels of indirection: we take the file index from
            # the LPE to grab the file_entry from the line program header,
            # then take the directory index from the file_entry to grab the
            # directory name from the line program header. Finally, we
            # join the (base) filename from the file_entry to the directory
            # name to get the absolute filename.
            lp_header = line_program.header
            file_entries = lp_header["file_entry"]

            # File and directory indices are 1-indexed.
            file_entry = file_entries[file_index] if line_program.header.version >= 5 else file_entries[file_index - 1]
            dir_index = file_entry["dir_index"] if line_program.header.version >= 5 else file_entry["dir_index"] - 1
            # dir_index = file_entry["dir_index"]
            assert dir_index >= 0

            # A dir_index of 0 indicates that no absolute directory was recorded during
            # compilation; return just the basename.
            if dir_index == 0:
                return file_entry.name.decode()

            # if min_dir_index == 0:
            if True:
                dir_index -= 1

            directory = lp_header["include_directory"][dir_index]
            # TODO: try out actual_path = op.normpath(CU.get_top_DIE().get_full_path())?
            return posixpath.join(directory, file_entry.name).decode()

        for CU in dwarfinfo.iter_CUs():
            line_program = dwarfinfo.line_program_for_CU(CU)
            if line_program is None:
                logger.warning("DWARF info is missing a line program for this CU")
                continue

            for DIE in CU.iter_DIEs():
                if (
                    DIE.tag == "DW_TAG_subprogram"
                    and "DW_AT_decl_file" in DIE.attributes
                    and "DW_AT_low_pc" in DIE.attributes
                    and "DW_AT_high_pc" in DIE.attributes
                ):
                    if "DW_AT_name" in DIE.attributes:
                        func_name = DIE.attributes["DW_AT_name"].value.decode()
                    else:
                        func_name = "???"
                    if "DW_AT_linkage_name" in DIE.attributes:
                        linkage_name = DIE.attributes["DW_AT_linkage_name"].value.decode()

                        unmangled_linkage_name = unmangle_helper(linkage_name)
                    else:
                        linkage_name = "???"
                        unmangled_linkage_name = "???"
                    if "DW_AT_decl_file" in DIE.attributes:
                        file_index = DIE.attributes["DW_AT_decl_file"].value
                        filename = lpe_filename(line_program, file_index)
                    else:
                        filename = "???"

                    srcFile_func_dict[filename][0].add(func_name)
                    srcFile_func_dict[filename][1].add(linkage_name)
                    srcFile_func_dict[filename][2].add(unmangled_linkage_name)

        for CU in dwarfinfo.iter_CUs():
            line_program = dwarfinfo.line_program_for_CU(CU)
            if line_program is None:
                logger.warning("  DWARF info is missing a line program for this CU")
                continue

            # CU_name = CU.get_top_DIE().attributes["DW_AT_name"].value.decode("utf-8")
            actual_path = os.path.normpath(CU.get_top_DIE().get_full_path())
            # print("actual_path", actual_path)

            for entry in line_program.get_entries():
                if entry.state:
                    pc = entry.state.address
                    line = entry.state.line
                    # print("line", line)
                    # pc_to_source_line_mapping[CU_name].append((pc, line))
                    pc_to_source_line_mapping[actual_path].append((pc, line))

            # if CU_name in pc_to_source_line_mapping:
            if actual_path in pc_to_source_line_mapping:
                # pc_to_source_line_mapping[CU_name].sort(key=lambda x: x[0])
                pc_to_source_line_mapping[actual_path].sort(key=lambda x: x[0])
    return func2pcs_data, srcFile_func_dict, pc_to_source_line_mapping


def analyze_dwarf(sess: Session, force: bool = False):
    logger.info("Analyzing DWARF info...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    # print("elf_artifacts", elf_artifacts)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]

    func2pc, file2funcs, file_pc2line = parse_dwarf(elf_artifact.path)
    file2funcs_data = [(file_name, vals[0], vals[1], vals[2]) for file_name, vals in file2funcs.items()]
    # print("func2pc", func2pc)
    # print("file2funcs", file2funcs)
    # print("file_func2line", file_pc2line)
    pc2locs = defaultdict(set)
    for file, pc_lines in file_pc2line.items():
        # print("file", file)
        for pc_line in list(set(pc_lines)):
            # print("pc_line", pc_line)
            assert len(pc_line) == 2
            pc, line = pc_line
            loc = f"{file}:{line}"
            pc2locs[pc].add(loc)
    func2pc_df = pd.DataFrame(func2pc, columns=["func", "pc_range"])
    file2funcs_df = pd.DataFrame(
        file2funcs_data,
        columns=["file", "func_names", "linkage_names", "unmangled_linkage_names"],
    )
    pc2locs_df = pd.DataFrame(pc2locs.items(), columns=["pc", "locs"])
    # print("func2pc_df", func2pc_df)
    # print("file2funcs_df", file2funcs_df)
    # print("pc2locs_df", pc2locs_df)

    attrs = {
        "elf_file": elf_artifact.name,
        "kind": "mapping",
        "by": "isaac_toolkit.analysis.static.dwarf",
    }

    func2pc_artifact = TableArtifact("func2pc", func2pc_df, attrs=attrs)
    file2funcs_artifact = TableArtifact("file2funcs", file2funcs_df, attrs=attrs)
    pc2locs_artifact = TableArtifact("pc2locs", pc2locs_df, attrs=attrs)
    # print("artifact1", func2pc_artifact)
    # print("artifact2", file2funcs_artifact)
    # print("artifact3", pc2locs_artifact)
    sess.add_artifact(func2pc_artifact, override=force)
    sess.add_artifact(file2funcs_artifact, override=force)
    sess.add_artifact(pc2locs_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    analyze_dwarf(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, List, Set, Tuple, Optional, Union

import bisect

import sys
import argparse
import posixpath
from pathlib import Path
from collections import defaultdict

from isaac_toolkit.session import Session
from isaac_toolkit.analysis.dynamic.trace.basic_blocks import BasicBlock  # TODO: move
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts
from isaac_toolkit.arch.riscv import riscv_branch_instrs, riscv_return_instrs
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.utils.demangle import unmangle_helper

logger = get_logger()


PC_FUNC_NAME_CACHE = {}


# TODO: reset?


def find_func_name(mapping: Dict[str, Tuple[int, int]], pc: int) -> str:
    # TODO: refactor to use existing tables
    """
    Given a program counter, find the function it belongs to
    """
    found = PC_FUNC_NAME_CACHE.get(pc)
    if found is not None:
        return found
    for func, ranges in mapping.items():
        for range_ in ranges:
            if pc >= range_[0] and pc <= range_[1]:
                PC_FUNC_NAME_CACHE[pc] = func
                return func
    ret = hex(pc)
    PC_FUNC_NAME_CACHE[pc] = ret
    return ret


def collect_bbs(trace_df, mapping):
    first_pc = None
    # TODO: make this generic!
    func2bbs = defaultdict(list)  # TODO: only track func_set?
    bb_freq = defaultdict(int)
    prev_pc = None
    prev_size = None
    # prev_instr = None
    bbs = []
    bb_instrs = []
    bb_size = 0
    trace_pcs = set(trace_df["pc"].unique())
    for row in trace_df.itertuples(index=False):
        pc = row.pc
        instr = row.instr
        instr = instr.strip()  # TODO: fix in frontend
        is_compressed = row.size == 2
        sz = 2 if is_compressed else 4

        if prev_pc:
            step = pc - prev_pc
            if step in [2, 4] and step == prev_size:
                pass
            else:
                if first_pc is None:
                    pass
                else:
                    logger.warning("Detected potential trap @ pc = 0x%x -> 0x%x", prev_pc, pc)
                    if True:
                        func = find_func_name(mapping, prev_pc)
                        bb = BasicBlock(
                            first_pc=first_pc,
                            last_pc=prev_pc,
                            num_instrs=len(bb_instrs),
                            size=bb_size,
                            end_instr=instr,
                            func=func,
                        )
                        bb_instrs = []
                        bb_size = 0
                        bbs.append(bb)
                        if bb.get_freq() == 1:
                            func2bbs[bb.func].append(bb)
                        bb_freq[bb] += 1

                        first_pc = pc

        # At the first pc of a basic block
        if first_pc is None:
            first_pc = pc

        bb_instrs.append(instr)
        bb_size += sz

        if instr in riscv_branch_instrs + riscv_return_instrs:
            func = find_func_name(mapping, pc)
            bb = BasicBlock(
                first_pc=first_pc,
                last_pc=pc,
                num_instrs=len(bb_instrs),
                size=bb_size,
                end_instr=instr,
                func=func,
            )
            bb_instrs = []
            bb_size = 0
            bbs.append(bb)
            if bb.get_freq() == 1:
                func2bbs[bb.func].append(bb)
                bb_freq[bb] += 1
            first_pc = None
        prev_pc = pc
        # prev_instr = instr
        prev_size = sz
    if first_pc is not None:
        func = None

    return bbs, trace_pcs, func2bbs, bb_freq


# TODO: consistent arg naming
# TODO: re-format quotes


def callgrind_format_get_inclusive_cost(bbs: List[BasicBlock]):
    call_stack = []
    bb_stack = []
    total_cost = 0
    inclusive_cost_dict = defaultdict(lambda: defaultdict(list))
    prev_bb = None

    # call_stack: [A, B]
    # [Given] bb_stack: [[bb1, bb2, bb3], [bb4, bb5, bb6]]
    # [B returns]: [[bb1, bb2, [bb3, # cost from B]]]
    # where bb1 - bb3 belong to func A and bb4 - bb6 belong to func B
    for i, bb in enumerate(bbs):
        total_cost += bb.num_instrs
        # print("bb", bb)
        # print("prev_bb", prev_bb)
        if prev_bb is None or (prev_bb.end_instr in riscv_branch_instrs and prev_bb.func != bb.func):
            # first bb in the trace
            call_stack.append(bb.func)
            bb_stack.append([bb])
        elif prev_bb.func == bb.func:
            # jalr doesn't necessarily mean return
            # 0x2ac8, jalr, memset -> 0x2ae8, memset
            bb_stack[-1].append(bb)
        # elif prev_bb.end_instr in return_instrs:
        # elif prev_bb.end_instr in riscv_return_instrs:
        elif prev_bb.end_instr in riscv_return_instrs or prev_bb.num_instrs == 1:
            # Check whether jalr refer to return
            # sometimes jalr simply means indirect jump
            # TODO: Redundant? Is it already handled in the above condition?
            if bb.func not in call_stack:
                call_stack.append(bb.func)
                bb_stack.append([bb])
                continue

            # ret case for the rest code snippet
            diff = len(call_stack) - call_stack.index(bb.func) - 1

            for j in range(diff):
                cost = 0
                callee_first_bb = bb_stack[-1][0][0] if isinstance(bb_stack[-1][0], list) else bb_stack[-1][0]
                for bb_stack_elem in bb_stack[-1]:
                    if not isinstance(bb_stack_elem, list):
                        # cost += 1 + (bb_stack_elem.last_pc - bb_stack_elem.first_pc) // 4
                        cost += bb_stack_elem.num_instrs
                    else:
                        subroutine_cost = bb_stack_elem[1]
                        cost += (
                            # subroutine_cost + 1 + (bb_stack_elem[0].last_pc - bb_stack_elem[0].first_pc) // 4
                            subroutine_cost
                            + bb_stack_elem[0].num_instrs
                        )

                bb_stack.pop()
                call_stack.pop()

                caller = bb_stack[-1][-1]
                if isinstance(caller, list):
                    caller_bb = caller[0]
                    inclusive_cost_dict[caller_bb.last_pc][callee_first_bb.first_pc].append(cost)
                    bb_stack[-1][-1][1] += cost
                else:
                    caller_bb = caller
                    inclusive_cost_dict[caller_bb.last_pc][callee_first_bb.first_pc].append(cost)
                    bb_stack[-1][-1] = [caller_bb, cost]

                # When callee function returns, it jumps to the start of the "next" basic block
                # in caller function. For example:
                # 0x6b0  jal    rd, imm <--- in function A
                # ...
                # 0x5ae0 jalr   rd, rs, imm <--- in function B
                # 0x6b4  ... <--- return to function A
                if j == diff - 1:
                    bb_stack[-1].append(bb)

        prev_bb = bb

    # print("inclusive_cost_dict", inclusive_cost_dict)
    # inclusive_cost_dict_sum = {pc: sum([sum(x) for x in costs.values()]) for pc, costs in inclusive_cost_dict.items()}
    # print("inclusive_cost_dict_sum", inclusive_cost_dict_sum)
    # inclusive_cost_dict_total = sum(inclusive_cost_dict_sum.values())
    # print("inclusive_cost_dict_total", inclusive_cost_dict_total)
    # inclusive_cost_dict_max = max(inclusive_cost_dict_sum.values())
    # print("inclusive_cost_dict_sum_max", inclusive_cost_dict_max)
    # print("total_cost", total_cost)
    # TODO: check that all costs are contained

    return inclusive_cost_dict


def callgrind_format_converter(
    bbs: List[BasicBlock],
    trace_pcs: Set[int],
    mapping: Dict[str, Tuple[int, int]],
    func2bbs: Dict[str, List[BasicBlock]],
    bb_freq: Dict[BasicBlock, int],
    srcFile_func_dict: Dict[str, List[str]],
    srcFile_linkage_name_dict: Dict[str, List[str]],
    srcFile_unmangled_linkage_name_dict: Dict[str, List[str]],
    func_set: Set[str],
    file2pc_loc: Optional[Dict[str, List[Tuple[int, int]]]] = None,
    dump_pc: bool = True,
    dump_pos: bool = False,
    elf_file_path: str = "/path/to/elf",
    unmangle_names: bool = False,
):
    """
    Collect necessary information for callgrind output format.
    Then dump it to output_file

    Necessary information is:
    1. function name and its source file name
    Output:
        ...
        fl={source_file_name}
        fn={func_name}

    2. If --dump-instr=yes
    Output:
        ...
        {pc} {frequency}
        ...

    3. If --dump-line=yes
    Output:
        ...
        {line_number_in_source_code} {frequency}
        ...

    4. Caller-callee relationship
    Output:
        ... in caller's context
        {pc} {frequency}
        cfi={callee's source file}
        calls={number of times caller calls callee} {callee pc}
        {pc} {inclusive_cost}
    """
    assert dump_pc or dump_pos
    inclusive_cost_dict = callgrind_format_get_inclusive_cost(bbs)

    # Find the basic block, of which the first pc is equal to pc
    # TODO: time complexity is suboptimal
    def find_bb(pc):
        for bb in bbs:
            if pc == bb.first_pc:
                return bb
        return None

    # Find the source file where the function is defined
    # Some functions come from libc. In this case ??? is returned.
    def find_srcFile(target_func: str) -> str:
        for srcFile, funcs in srcFile_func_dict.items():
            for func in funcs:
                if func == target_func:
                    return srcFile
        for srcFile, funcs in srcFile_linkage_name_dict.items():
            for func in funcs:
                if func == target_func:
                    return srcFile
        return "???"

    #
    def aggregate_pos_cost_of_func(sorted_bb_lists: list):
        position_cost_dict = defaultdict(int)
        for bb in sorted_bb_lists:
            pc = bb.first_pc
            # while pc <= bb.last_pc:
            for pc_ in range(pc, bb.last_pc + 2, 2):
                if pc_ not in trace_pcs:
                    # pc += 2
                    continue
                position_cost_dict[pc_] += bb.get_freq()
                # pc += 4
                # pc += 2
        return position_cost_dict

    positions = "instr" if dump_pc else "line"

    prologue = f"""\
# callgrind format
version: 1
creator: callgrind-3.15.0
pid: 2577934
cmd:  ./dhrystone
part: 1

desc: I1 cache:
desc: D1 cache:
desc: LL cache:

desc: Timerange:
desc: Trigger: Program termination

positions: {positions}
events: Ir
summary:

"""
    callgrind_output = ""

    def callgrind_format_dump_instr(pc: int, source_file: str) -> str:
        return hex(pc)

    def callgrind_format_dump_line(pc: int, source_file: str) -> str:
        if source_file == "???":
            return "0"
        assert file2pc_loc is not None
        mapping_ = file2pc_loc.get(source_file)
        if mapping_ is None:
            return "0"
        # PYTHON 3.10:
        # i = bisect.bisect_right(mapping_, pc, key=lambda x: x[0])
        # if i:
        #     start_pc, line = mapping_[i - 1]
        #     if pc >= start_pc:
        #         return f"{line}"
        # ---
        # PYTHON 3.8
        mapping_first = [x[0] for x in mapping_]
        i = bisect.bisect_right(mapping_first, pc)
        if i:
            start_pc, line = mapping_[i - 1]
            if pc >= start_pc:
                return f"{line}"
        # ---
        return "0"

    dump_positions = callgrind_format_dump_instr if dump_pc else callgrind_format_dump_line

    # source file to functions mapping
    srcFile_to_func = defaultdict(list)
    for func in func_set:
        srcFile_to_func[find_srcFile(func)].append(func)

    for srcFile, funcs in srcFile_to_func.items():
        callgrind_output += f"ob={posixpath.abspath(elf_file_path)}\n"
        callgrind_output += f"fl={srcFile}\n"

        for func in funcs:

            bb_list = func2bbs[func]
            bb_list.sort(key=lambda bb: bb.first_pc)
            if unmangle_names:
                func = unmangle_helper(func)
            callgrind_output += f"fn={func}\n"

            position_cost_dict = aggregate_pos_cost_of_func(bb_list)

            branch_pc_list = [bb.last_pc for bb in bb_list]

            for pc in sorted(position_cost_dict.keys()):
                position_info = dump_positions(pc, srcFile)
                callgrind_output += f"{position_info} {position_cost_dict[pc]}\n"
                if pc in branch_pc_list and pc in inclusive_cost_dict:
                    for callee_pc, inclusive_cost in inclusive_cost_dict[pc].items():
                        # TODO Share object files case not implemented here
                        callgrind_output += f"cob={posixpath.abspath(elf_file_path)}\n"
                        callee_func = find_func_name(mapping, callee_pc)
                        callgrind_output += f"cfi={find_srcFile(callee_func)}\n"
                        if unmangle_names:
                            callee_func = unmangle_helper(callee_func)
                        callgrind_output += f"cfn={callee_func}\n"
                        callgrind_output += f"calls={len(inclusive_cost)} {hex(callee_pc)}\n"
                        callgrind_output += f"{position_info} {sum(inclusive_cost)}\n"

            callgrind_output += "\n"

    content = prologue + callgrind_output
    return content


def generate_callgrind_output(
    sess: Session,
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
    dump_pc: bool = False,
    dump_pos: bool = False,
    unmangle_names: bool = False,
):
    logger.info("Generating callgrind coutput...")
    assert output is not None
    artifacts = sess.artifacts
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]

    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]

    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.name == "func2pc")
    assert len(func2pc_artifacts) == 1
    func2pc_artifact = func2pc_artifacts[0]
    func2pc_df = func2pc_artifact.df

    file2funcs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "file2funcs")
    assert len(file2funcs_artifacts) == 1
    file2funcs_artifact = file2funcs_artifacts[0]
    file2funcs_df = file2funcs_artifact.df

    pc2locs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "pc2locs")
    assert len(pc2locs_artifacts) == 1
    pc2locs_artifact = pc2locs_artifacts[0]
    pc2locs_df = pc2locs_artifact.df

    def helper(pc2locs_df):
        df_ = pc2locs_df.explode("locs")
        df_.reset_index(inplace=True)
        df_[["file", "line"]] = df_["locs"].str.split(":", n=1, expand=True)
        ret = {}
        for file, file_df in df_.groupby("file"):
            temp = list(map(tuple, file_df[["pc", "line"]].values))
            temp = sorted(temp, key=lambda x: x[0])
            ret[file] = temp
        return ret

    file2pc_loc = helper(pc2locs_df)

    mapping = func2pc_df.groupby("func")["pc_range"].apply(list).to_dict()
    bbs, trace_pcs, func2bbs, bb_freq = collect_bbs(trace_artifact.df, mapping)
    file2funcs = dict(list((file2funcs_df[["file", "func_names"]].to_records(index=False))))
    file2linkage_names = dict(list((file2funcs_df[["file", "linkage_names"]].to_records(index=False))))
    file2unmangled_linkage_names = dict(
        list((file2funcs_df[["file", "unmangled_linkage_names"]].to_records(index=False)))
    )
    func_set = set(func2bbs.keys())
    elf_file_path = elf_artifact.path

    content = callgrind_format_converter(
        bbs=bbs,
        trace_pcs=trace_pcs,
        mapping=mapping,
        func2bbs=func2bbs,
        bb_freq=bb_freq,
        srcFile_func_dict=file2funcs,
        srcFile_linkage_name_dict=file2linkage_names,
        srcFile_unmangled_linkage_name_dict=file2unmangled_linkage_names,
        func_set=func_set,
        file2pc_loc=file2pc_loc,
        dump_pc=dump_pc,
        dump_pos=dump_pos,
        elf_file_path=elf_file_path,
        unmangle_names=unmangle_names,
    )

    if output is None:
        profile_dir = sess.directory / "profile"
        profile_dir.mkdir(exist_ok=True)
        out_name = "callgrind"
        if dump_pc:
            out_name += "_pc"
        if dump_pos:
            out_name += "_pos"
        out_name += ".out"
        output = profile_dir / out_name
    with open(output, "w") as f:
        f.write(content)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    generate_callgrind_output(
        sess,
        output=args.output,
        force=args.force,
        dump_pc=args.dump_pc,
        dump_pos=args.dump_pos,
        unmangle_names=args.unmangle,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--output", default=None)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--dump-pc", action="store_true")
    parser.add_argument("--dump-pos", action="store_true")
    parser.add_argument("--unmangle", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

import yaml

from .config import IsaacConfig, DEFAULT_CONFIG, ArtifactsSettings, TableArtifactsSettings
from .artifact import (
    FileArtifact,
    ElfArtifact,
    InstrTraceArtifact,  # TODO: drop
    TraceArtifact,
    SourceArtifact,
    DisassArtifact,
    TableArtifact,
    M2ISARArtifact,
    GraphArtifact,
    ArtifactFlag,
    PythonArtifact,
    filter_artifacts,
)
from isaac_toolkit.logging import get_logger, set_log_level, set_log_file

logger = get_logger()


# def create_dirs(base, dirnames):
#     assert isinstance(base, Path)
#     for dirname in dirnames:
#         (base / dirname).mkdir()

COMPRESSION_EXT = {
    "zip": "zip",
    # "gzip": ".gz",
    "gzip": "gz",
    "bz2": "bz2",
    "zstd": "zst",
    "xz": "tar.xz",
    "tar": "tar",
}

FMT_EXT = {
    "pickle": "pkl",
    "parquet": "parquet",
}


def get_table_artifact_ext(table_artifacts_settings: TableArtifactsSettings):

    fmt = table_artifacts_settings.fmt
    # engine = table_artifacts_settings.engine
    compression_method = table_artifacts_settings.compression_method
    # compression_level = table_artifacts_settings.compression_level

    fmt_ext = FMT_EXT.get(fmt, None)
    assert fmt_ext is not None, f"Unsupported fmt: {fmt}"
    ret = fmt_ext

    compression_ext = None
    if compression_method is not None and compression_method != "none":
        compression_ext = COMPRESSION_EXT.get(compression_method, compression_method)
    if compression_ext:
        ret += f".{compression_ext}"
    return ret


def get_artifact_ext(flags: ArtifactFlag, artifacts_settings: ArtifactsSettings):
    if flags & ArtifactFlag.ELF:
        ext = None
    elif flags & ArtifactFlag.INSTR_TRACE:
        ext = get_table_artifact_ext(artifacts_settings.instr_trace)
    elif flags & ArtifactFlag.TRACE:
        ext = get_table_artifact_ext(artifacts_settings.trace)
    elif flags & (ArtifactFlag.SOURCE | ArtifactFlag.DISASS):
        ext = None
    elif flags & ArtifactFlag.GRAPH:
        ext = "pkl"
    elif flags & ArtifactFlag.TABLE:
        ext = get_table_artifact_ext(artifacts_settings.table)
    elif flags & ArtifactFlag.M2ISAR:
        ext = "m2isarmodel"
    elif flags & ArtifactFlag.PYTHON:
        ext = "pkl"
    else:
        raise RuntimeError("Unhandled case!")
    return ext


def load_artifacts(base):
    artifacts_yaml = base / "artifacts.yml"
    if not artifacts_yaml.is_file():
        return []
    with open(artifacts_yaml, "r", encoding="utf-8") as f:
        yaml_data = yaml.safe_load(f)
    artifacts = yaml_data.get("artifacts", None)
    assert artifacts is not None
    artifacts_ = []
    for artifact in artifacts:
        name = artifact.get("name", None)
        assert name is not None
        dest = artifact.get("dest", None)
        assert dest is not None
        flags = artifact.get("flags", None)
        assert flags is not None
        attrs = artifact.get("attrs", None)
        assert attrs is not None
        flags_ = ArtifactFlag(flags)
        # print("flags_", flags)
        # if flags_ & (ArtifactFlag.INPUT | ArtifactFlag.OUTPUT | ArtifactFlag.TEMP):
        #     path = artifact.get("path", None)
        #     assert path is not None
        if flags_ & ArtifactFlag.ELF:
            artifact_ = ElfArtifact.from_dict(artifact)
            # (name, dest, flags=flags, attrs=attrs)
        elif flags_ & ArtifactFlag.INSTR_TRACE:  # TODO: drop
            artifact_ = InstrTraceArtifact.from_dict(artifact)
        elif flags_ & ArtifactFlag.TRACE:
            artifact_ = TraceArtifact.from_dict(artifact)
        elif flags_ & ArtifactFlag.SOURCE:
            artifact_ = SourceArtifact.from_dict(artifact)
        elif flags_ & ArtifactFlag.DISASS:
            artifact_ = DisassArtifact.from_dict(artifact)
        elif flags_ & ArtifactFlag.GRAPH:
            artifact_ = GraphArtifact.from_dict(artifact)
        elif flags_ & ArtifactFlag.TABLE:
            artifact_ = TableArtifact.from_dict(artifact)
        elif flags_ & ArtifactFlag.M2ISAR:
            artifact_ = M2ISARArtifact.from_dict(artifact)
        elif flags_ & ArtifactFlag.PYTHON:
            artifact_ = PythonArtifact.from_dict(artifact)
        else:
            logger.warning("Unhandled artifact type!")
            artifact_ = FileArtifact.from_dict(artifact)
            raise RuntimeError("Unhandled case!")
        artifacts_.append(artifact_)
    # TODO: check for duplicates
    # print("artifacts", artifacts)
    # print("artifacts_", artifacts_)
    return artifacts_


class Session:

    def __init__(self, session_dir: Path, config: IsaacConfig):
        self.directory = session_dir.resolve()
        self.config = config
        self._artifacts = load_artifacts(self.directory)

    @property
    def artifacts(self):
        return self._artifacts

    def add_artifact(self, artifact, override=False):
        logger.debug("Adding artifact to session")
        artifact_names = [x.name for x in self._artifacts]
        if artifact.name in artifact_names:
            if override:
                logger.debug("Overriding artifact")
                idx = artifact_names.index(artifact.name)
                del self._artifacts[idx]
            else:
                raise RuntimeError(
                    f"Artifact with name {artifact.name} already exists. Use override=True or cleanup session."
                )
        self._artifacts.append(artifact)

    # @property
    # def inputs(self):
    #     return filter_artifacts(self.artifacts, lambda x: ArtifactFlag.INPUT & x.flags)

    # @property
    # def outputs(self):
    #     return filter_artifacts(self.artifacts, lambda x: ArtifactFlag.OUTPUT & x.flags)

    # @property
    # def temps(self):
    #     return filter_artifacts(self.artifacts, lambda x: ArtifactFlag.TEMP & x.flags)

    @property
    def graphs(self):
        return filter_artifacts(self.artifacts, lambda x: ArtifactFlag.GRAPH & x.flags)

    @property
    def tables(self):
        return filter_artifacts(self.artifacts, lambda x: ArtifactFlag.TABLE & x.flags)

    # @property
    # def temp_dir(self):
    #     return self.directory / "temp"

    # @property
    # def inputs_dir(self):
    #     return self.directory / "inputs"

    # @property
    # def outputs_dir(self):
    #     return self.directory / "outputs"

    # @property
    # def graphs_dir(self):
    #     return self.directory / "graphs"

    # @property
    # def tables_dir(self):
    #     return self.directory / "tables"

    def validate(self):
        pass
        # assert self.temp_dir.is_dir()
        # assert self.inputs_dir.is_dir()
        # assert self.outputs_dir.is_dir()
        # self.config.validate()

    def save_artifacts(self):
        # print("save_artifacts")
        logger.debug("Saving artifacts...")
        artifacts_ = []
        for artifact in self.artifacts:
            dest_dir = None
            dest_file = artifact.name
            ext = get_artifact_ext(artifact.flags, self.config.artifacts)
            if ext is not None and len(ext) > 0:
                dest_file = f"{dest_file}.{ext}"
            if isinstance(artifact, ElfArtifact):
                dest_dir = self.directory / "elf"
            elif isinstance(artifact, InstrTraceArtifact):
                dest_dir = self.directory / "instr_trace"
            elif isinstance(artifact, TraceArtifact):
                dest_dir = self.directory / "trace"
            elif isinstance(artifact, SourceArtifact):
                dest_dir = self.directory / "source"
            elif isinstance(artifact, GraphArtifact):
                # assert not artifact.is_input and not artifact.is_output
                dest_dir = self.directory / "graph"
            elif isinstance(artifact, TableArtifact):
                # assert not artifact.is_input and not artifact.is_output
                dest_dir = self.directory / "table"
            elif isinstance(artifact, M2ISARArtifact):
                # assert not artifact.is_input and not artifact.is_output
                dest_dir = self.directory / "model"
            elif isinstance(artifact, PythonArtifact):
                # assert not artifact.is_input and not artifact.is_output
                dest_dir = self.directory / "misc"
            if dest_dir is None:
                dest_dir = self.directory / "misc"
            assert dest_file is not None
            assert len(dest_file) > 0
            assert dest_file[0] != "/"
            dest = dest_dir / dest_file
            dest.parent.mkdir(parents=True, exist_ok=True)
            artifact.save(dest, self.config.artifacts)
            metadata = artifact.to_dict()
            metadata["dest"] = str(dest)
            artifacts_.append(metadata)
        yaml_data = {"artifacts": artifacts_}
        artifacts_yaml = self.directory / "artifacts.yml"
        with open(artifacts_yaml, "w") as f:
            yaml.dump(yaml_data, f)

    def save(self):
        # print("save")
        self.config.to_yaml_file(self.directory / "config.yml")
        self.save_artifacts()

    @staticmethod
    def create(session_dir):
        if isinstance(session_dir, str):
            session_dir = Path(session_dir)
        assert isinstance(session_dir, Path)
        assert session_dir.parent.is_dir(), f"Parent directory does not exist: {session_dir.parent}"
        session_dir.mkdir()
        # create_dirs(session_dir, ["inputs", "outputs", "temp", "graphs", "tables"])
        config = IsaacConfig.from_dict(DEFAULT_CONFIG)
        sess = Session(session_dir, config)
        log_file = session_dir / "session.log"
        set_log_file(log_file)
        sess.save()
        return sess

    @staticmethod
    def from_dir(session_dir):
        if isinstance(session_dir, str):
            session_dir = Path(session_dir)
        assert isinstance(session_dir, Path)
        assert session_dir.is_dir(), f"Directory does not exist: {session_dir}"
        config_file = session_dir / "config.yml"
        assert config_file.is_file(), f"Missing config file: {config_file}"
        config = IsaacConfig.from_yaml_file(config_file)
        log_file = session_dir / "session.log"
        set_log_file(log_file)
        set_log_level(console_level=config.logging.console.level, file_level=config.logging.file.level)
        sess = Session(session_dir, config)
        sess.validate()
        sess.save()
        return sess



import sys
import argparse

import pandas as pd
import yaml


def annotate_enc_score(index, enc_score_csv, output=None, inplace=False):

    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)
    candidates = combined_index_data["candidates"]
    # print("candidates", candidates)

    scores_df = pd.read_csv(enc_score_csv)
    # print("scores_df", scores_df)
    assert len(candidates) == len(scores_df)

    for i, candidate in enumerate(candidates):
        # print("i", i)
        name = candidate["properties"]["InstrName"]
        instr_row = scores_df[scores_df["instr"] == name]
        assert len(instr_row) == 1
        enc_score = float(instr_row["enc_score"].iloc[0])
        enc_weight = float(instr_row["weight"].iloc[0])
        metrics = candidate.get("metrics", {})
        # print("metrics", metrics)
        metrics["enc_weight"] = enc_weight
        metrics["enc_score"] = enc_score
        # print("metrics2", metrics)
        candidate["metrics"] = metrics

    if inplace:
        assert output is None
        out_file = index
    else:
        assert output is not None
        out_file = output

    combined_index_data["candidates"] = candidates

    with open(out_file, "w") as f:
        yaml.dump(combined_index_data, f)


def handle(args):
    annotate_enc_score(args.index, args.enc_score_csv, output=args.output, inplace=args.inplace)


def get_parser():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument("--inplace", action="store_true", help="TODO")
    parser.add_argument("--enc-score-csv", required=True, help="TODO")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional
from pathlib import Path

from isaac_toolkit.session import Session

# from isaac_toolkit.logging import get_logger, set_log_level

from .stage.load import load_artifacts
from .stage.normalize import normalize_artifacts
from .stage.analyze import analyze_artifacts
from .stage.visualize import visualize_artifacts
from .stage.report import generate_reports
from .stage.profile import generate_profile
from .stage.lcov import generate_lcov

# logger = get_logger()
import logging

logger = logging.getLogger()


def run_full_flow(
    sess: Session,
    elf_file: Optional[Path] = None,
    linker_map_file: Optional[Path] = None,
    instr_trace_file: Optional[Path] = None,
    disass_file: Optional[Path] = None,
    unmangle: bool = False,
    force: bool = False,
    progress: bool = False,
    report_fmt="md",
    report_detailed=False,
    report_portable=False,
    report_style=False,
    report_topk=10,
):
    logger.info("Running full RVF flow...")
    load_artifacts(
        sess,
        elf_file=elf_file,
        linker_map_file=linker_map_file,
        instr_trace_file=instr_trace_file,
        disass_file=disass_file,
        force=force,
    )
    normalize_artifacts(sess, force=force)
    analyze_artifacts(sess, force=force)
    visualize_artifacts(sess, force=force)
    generate_reports(
        sess,
        fmt=report_fmt,
        detailed=report_detailed,
        portable=report_portable,
        style=report_style,
        topk=report_topk,
        force=force,
    )
    generate_profile(sess, force=force, unmangle=unmangle)
    generate_lcov(sess, force=force, unmangle=unmangle)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)

    # set_log_level(console_level=args.log, file_level=args.log)
    def path_helper(x):
        if x is None:
            return x
        return Path(x)

    run_full_flow(
        sess,
        elf_file=path_helper(args.elf),
        linker_map_file=path_helper(args.linker_map),
        instr_trace_file=path_helper(args.instr_trace),
        disass_file=path_helper(args.disass),
        force=args.force,
        unmangle=args.unmangle,
        report_fmt=args.report_fmt,
        report_detailed=args.report_detailed,
        report_portable=args.report_portable,
        report_style=args.report_style,
        report_topk=args.report_topk,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--progress", action="store_true")
    parser.add_argument("--elf", default=None)
    parser.add_argument("--linker-map", default=None)
    parser.add_argument("--instr-trace", default=None)
    parser.add_argument("--disass", default=None)
    parser.add_argument("--unmangle", action="store_true")
    parser.add_argument("--report-fmt", choices=["md", "txt", "html", "pdf"], default="md", help="Report format")
    parser.add_argument("--report-detailed", action="store_true", help="Include detailed breakdowns")
    parser.add_argument("--report-portable", action="store_true", help="Embed plots as base64 (only for HTML output)")
    parser.add_argument("--report-style", action="store_true", help="Use custom CSS for HTML output")
    parser.add_argument("--report-topk", type=int, default=10, help="Limit number of table rows")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse

import yaml


def sort_index(index, output=None, inplace=False, ascending=False, by=None, by2=None):
    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)
    candidates = combined_index_data["candidates"]
    print("candidates", candidates)

    candidates = sorted(
        candidates, key=lambda x: (x["metrics"].get(by, 0.0), x["metrics"].get(by2, 0.0)), reverse=not ascending
    )
    combined_index_data["candidates"] = candidates
    print("sorted_candidates", candidates)

    if inplace:
        assert output is None
        out_file = index
    else:
        assert output is not None
        out_file = output

    with open(out_file, "w") as f:
        yaml.dump(combined_index_data, f)


def handle(args):
    sort_index(args.index, output=args.output, inplace=args.inplace, ascending=args.ascending, by=args.by, by2=args.by2)


def get_parser():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument("--inplace", action="store_true", help="TODO")
    parser.add_argument("--ascending", action="store_true", help="TODO")
    parser.add_argument("--by", required=True, help="Metric used for sorting")
    parser.add_argument("--by2", default=None, help="Metric used for sorting 2nd level")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from .opcode import decode_opcode
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def create_opcode_per_llvm_bb_hist(sess: Session, force: bool = False):
    logger.info("Analyzing opcodes per LLVM BB...")
    artifacts = sess.artifacts
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]
    trace_df = trace_artifact.df

    llvm_bbs_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs"
    )  # TODO: optional or different pass
    assert len(llvm_bbs_artifacts) == 1
    llvm_bbs_artifact = llvm_bbs_artifacts[0]
    llvm_bbs_df = llvm_bbs_artifact.df.copy()
    llvm_bbs_df[["start", "end"]] = llvm_bbs_df["pcs"].apply(pd.Series)
    # print("llvm_bbs_df", llvm_bbs_df)

    # TODO: use pc_hist
    pc_counts = trace_df["pc"].value_counts().rename("count").reset_index()
    pc_counts.columns = ["pc", "count"]
    total_count = pc_counts["count"].sum()

    # extract table of unique (pc, bytecode)
    unique_pc_df = trace_df.drop_duplicates(subset=["pc"])[["pc", "bytecode"]]

    # apply detect_opcode once per PC
    unique_pc_df["opcode"] = unique_pc_df["bytecode"].apply(decode_opcode)

    # keep only (pc, opcode)
    pc_to_opcode = unique_pc_df[["pc", "opcode"]]

    pc_opcode_df = pc_counts.merge(pc_to_opcode, on="pc", how="left").sort_values("pc").reset_index(drop=True)

    dfs = []
    for _, row in llvm_bbs_df.iterrows():
        start = row["start"]
        end = row["end"]

        # Efficient filter: this DataFrame is tiny compared to trace_df
        slice_df = pc_opcode_df.loc[(pc_opcode_df["pc"] >= start) & (pc_opcode_df["pc"] < end)]

        if slice_df.empty:
            continue

        # aggregate counts by opcode
        op_hist = slice_df.groupby("opcode")["count"].sum().reset_index()

        op_hist.insert(0, "func_name", row["func_name"])
        op_hist.insert(1, "bb_name", row["bb_name"])

        dfs.append(op_hist)

    df = pd.concat(dfs)
    df["rel_count"] = df["count"] / total_count

    attrs = {
        "kind": "histogram",
        "by": __name__,
    }

    opcodes_artifact = TableArtifact("opcodes_per_llvm_bb_hist", df, attrs=attrs)
    sess.add_artifact(opcodes_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    create_opcode_per_llvm_bb_hist(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional, List
from pathlib import Path

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.retargeting.llvm.seal5 import retarget_seal5_llvm

logger = get_logger()


def retarget_llvm_auto(
    sess: Session,
    cfg_files: List[str],
    label: str = None,
    workdir: Optional[str] = None,
    splitted: bool = False,
    force: bool = False,
    progress: bool = False,
    docker_override: Optional[bool] = None,
    verbose: bool = False,  # TODO: add to other stages, too
    cleanup: bool = False,
):
    logger.info("Retargeting LLVM with ISAAC instructions...")
    cfg_files = [Path(f) for f in cfg_files]
    assert all(f.is_file() for f in cfg_files)
    # input(">>>")
    tool = "seal5"  # TODO: read from config
    assert tool == "seal5"
    tool_retarget_llvm = {
        "seal5": retarget_seal5_llvm,
    }
    retarget_llvm = tool_retarget_llvm.get(tool)
    assert retarget_llvm is not None
    if workdir is None:
        workdir = sess.directory / "work"  # expose via cmd
    else:
        workdir = Path(workdir)
        assert workdir.is_dir()
    config = sess.config
    flow_config = config.flow
    assert flow_config is not None
    demo_config = flow_config.demo
    assert demo_config is not None
    docker_config = demo_config.docker
    assert docker_config is not None
    use_docker = docker_override if docker_override is not None else docker_config.enable
    riscv_config = demo_config.riscv
    assert riscv_config is not None
    coredsl_config = demo_config.coredsl
    assert coredsl_config is not None
    mount_dir = None  # Expose to CLI
    set_name = coredsl_config.set_name
    xlen = riscv_config.xlen
    seal5_sets = []
    # if splitted:
    if False:
        raise NotImplementedError("Splitted")
    else:
        seal5_sets.append(set_name)
    if use_docker:
        docker_image = docker_config.seal5_image
        if mount_dir is None:
            mount_dir = Path.cwd()
    else:
        docker_image = None
    retarget_llvm(
        sess,
        workdir=workdir,
        mount_dir=mount_dir,
        docker_image=docker_image,
        seal5_sets=seal5_sets,
        xlen=xlen,
        cfg_files=cfg_files,
        label=label,
        force=force,
        verbose=verbose,
        cleanup=cleanup,
        progress=progress,
        splitted=splitted,
    )


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    docker_override = None
    if args.docker:
        assert not args.local
        docker_override = True
    if args.local:
        assert not args.docker
        docker_override = False
    retarget_llvm_auto(
        sess,
        args.cfg_file,
        label=args.label,
        workdir=args.workdir,
        splitted=args.splitted,
        force=args.force,
        docker_override=docker_override,
        verbose=args.verbose,
        cleanup=args.cleanup,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("cfg_file", nargs="+")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--label", default=None)
    parser.add_argument("--workdir", default=None)
    parser.add_argument("--splitted", action="store_true")
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--docker", action="store_true")
    parser.add_argument("--local", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("--cleanup", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import argparse
from typing import Dict, Optional
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import yaml


def main():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument("--inplace", action="store_true", help="TODO")
    parser.add_argument("--hls-schedules-csv", required=True, help="TODO")
    parser.add_argument("--hls-selected-schedules-yaml", required=True, help="TODO")
    args = parser.parse_args()

    with open(args.index, "r") as f:
        combined_index_data = yaml.safe_load(f)
    candidates = combined_index_data["candidates"]
    # print("candidates", candidates)
    global_data = combined_index_data["global"]
    # print("global_data", global_data)
    global_artifacts = global_data["artifacts"]
    hls_dir = Path(args.hls_schedules_csv).parent
    global_artifacts["HLS_DIR"] = str(hls_dir)
    global_data["artifacts"] = global_artifacts
    combined_index_data["global"] = global_data

    schedules_df = pd.read_csv(args.hls_schedules_csv)
    schedules_df
    # print("schedules_df", schedules_df)
    # TODO: each instr needs its own schedule!

    with open(args.hls_selected_schedules_yaml, "r") as f:
        yaml_data = yaml.safe_load(f)
    # print("yaml_data", yaml_data)
    instr2ii = {}
    instr2lat = {}
    instr2area = {}
    instr2group = {}
    instr2num_shared = {}
    group2instrs = {}
    for row in yaml_data:
        sharing_group = row["sharing_group"]
        idx = row["solution_idx"]
        name = f"SG_{sharing_group}_SOL_IDX_{idx}"
        schedules = schedules_df[schedules_df["config"] == name]
        assert len(schedules) == 1
        # print("schedules", schedules)
        schedule = schedules.iloc[0]
        ii = int(schedule["II"])
        import ast

        lats = ast.literal_eval(schedule["Instruction latencies"])
        area = float(schedule["Area estimate w/o lifetimes"])
        instrs = list(lats.keys())
        group2instrs[idx] = instrs
        for instr, lat in lats.items():
            instr2ii[instr] = ii
            instr2lat[instr] = lat
            instr2area[instr] = area
            instr2group[instr] = sharing_group
            instr2num_shared[instr] = len(lats)
    assert len(candidates) == len(instr2ii)

    for i, candidate in enumerate(candidates):
        name = candidate["properties"]["InstrName"]
        metrics = candidate.get("metrics", {})
        metrics["hls_ii"] = instr2ii[name]
        metrics["hls_latency"] = instr2lat[name]
        metrics["hls_area"] = instr2area[name]
        metrics["hls_num_shared"] = instr2num_shared[name]
        area_scaled = instr2area[name] / instr2num_shared[name]
        metrics["hls_area_scaled"] = area_scaled
        # print("metrics", metrics)
        candidate["metrics"] = metrics

    if args.inplace:
        assert args.output is None
        out_file = args.index
    else:
        assert args.output is not None
        out_file = args.output

    combined_index_data["candidates"] = candidates

    with open(out_file, "w") as f:
        yaml.dump(combined_index_data, f)


if __name__ == "__main__":
    main()
